"""Zemel algorithm."""
from __future__ import annotations

import json
import sys
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple, Tuple

import numpy as np
import pandas as pd
import scipy.optimize as optim
from joblib import dump, load
from scipy.spatial.distance import cdist
from scipy.special import softmax  # type: ignore[attr-defined]

from ethicml.implementations.utils import load_data_from_flags, save_transformations
from ethicml.utility import DataTuple, TestTuple

if TYPE_CHECKING:
    from ethicml.algorithms.preprocess.pre_subprocess import PreAlgoArgs, T
    from ethicml.algorithms.preprocess.zemel import ZemelArgs


class Model(NamedTuple):
    """Model."""

    prototypes: np.ndarray
    w: np.ndarray


def LFR_optim_objective(
    parameters: np.ndarray,
    x_unprivileged: np.ndarray,
    x_privileged: np.ndarray,
    y_unprivileged: np.ndarray,
    y_privileged: np.ndarray,
    clusters: int,
    A_x: float,
    A_y: float,
    A_z: float,
    print_interval: int,
    verbose: bool,
) -> np.number:
    """LFR optim objective.

    :param parameters:
    :param x_unprivileged:
    :param x_privileged:
    :param y_unprivileged:
    :param y_privileged:
    :param clusters:
    :param A_x:
    :param A_y:
    :param A_z:
    :param print_interval:
    :param verbose:
    """
    _, features_dim = x_unprivileged.shape

    w = parameters[:clusters]
    prototypes = parameters[clusters:].reshape((clusters, features_dim))

    m_unprivileged, x_hat_unprivileged, y_hat_unprivileged = get_xhat_y_hat(
        prototypes, w, x_unprivileged
    )

    m_privileged, x_hat_privileged, y_hat_privileged = get_xhat_y_hat(prototypes, w, x_privileged)

    y_hat = np.concatenate([y_hat_unprivileged, y_hat_privileged], axis=0)
    y = np.concatenate([y_unprivileged.reshape((-1, 1)), y_privileged.reshape((-1, 1))], axis=0)

    l_x = np.mean((x_hat_unprivileged - x_unprivileged) ** 2) + np.mean(
        (x_hat_privileged - x_privileged) ** 2
    )
    l_z = np.mean(abs(np.mean(m_unprivileged, axis=0) - np.mean(m_privileged, axis=0)))
    l_y = -np.mean(y * np.log(y_hat) + (1.0 - y) * np.log(1.0 - y_hat))

    total_loss = A_x * l_x + A_y * l_y + A_z * l_z

    if verbose and LFR_optim_objective.steps % print_interval == 0:  # type: ignore[attr-defined]
        print(
            f"step: {LFR_optim_objective.steps}, "  # type: ignore[attr-defined]
            f"loss: {total_loss}, "
            f"L_x: {l_x},  "
            f"L_y: {l_y},  "
            f"L_z: {l_z}"
        )
    LFR_optim_objective.steps += 1  # type: ignore[attr-defined]

    return total_loss


def get_xhat_y_hat(
    prototypes: np.ndarray, w: np.ndarray, x: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Get xhat y hat.

    :param prototypes:
    :param w:
    :param x:
    """
    M = softmax(-cdist(x, prototypes), axis=1)
    x_hat = np.matmul(M, prototypes)
    y_hat = np.clip(
        np.matmul(M, w.reshape((-1, 1))), np.finfo(float).eps, 1.0 - np.finfo(float).eps
    )
    return M, x_hat, y_hat


def train_and_transform(
    train: DataTuple, test: T, flags: ZemelArgs, seed: int
) -> (Tuple[DataTuple, T]):
    """Train and transform.

    :param train:
    :param test:
    :param flags:
    """
    prototypes, w = fit(train, flags, seed=seed)

    training_sensitive = train.x.loc[train.s == 0].to_numpy()
    training_nonsensitive = train.x.loc[train.s == 1].to_numpy()

    testing_sensitive = test.x.loc[test.s == 0].to_numpy()
    testing_nonsensitive = test.x.loc[test.s == 1].to_numpy()

    train_transformed = trans(prototypes, w, training_nonsensitive, training_sensitive, train)
    test_transformed = trans(prototypes, w, testing_nonsensitive, testing_sensitive, test)

    return train.replace(x=train_transformed), test.replace(x=test_transformed)


def transform(data: T, prototypes: np.ndarray, w: np.ndarray) -> T:
    """Transform.

    :param data:
    :param prototypes:
    :param w:
    """
    data_sens = data.x.loc[data.s == 0].to_numpy()
    data_nons = data.x.loc[data.s == 1].to_numpy()
    transformed = trans(prototypes, w, data_nons, data_sens, data)
    return data.replace(x=transformed)


def fit(train: DataTuple, flags: ZemelArgs, seed: int) -> Model:
    """Train the Zemel model and return the transformed features of the train and test sets.

    :param train:
    :param flags:
    """
    np.random.seed(seed)

    training_sensitive = train.x.loc[train.s == 0].to_numpy()
    training_nonsensitive = train.x.loc[train.s == 1].to_numpy()
    ytrain_sensitive = train.y.loc[train.s == 0].to_numpy()
    ytrain_nonsensitive = train.y.loc[train.s == 1].to_numpy()

    print_interval = 100
    verbose = False

    _, features_dim = train.x.shape

    # Initialize the LFR optim objective parameters
    parameters_initialization = np.random.uniform(
        size=flags["clusters"] + features_dim * flags["clusters"]
    )
    bnd = [(0, 1)] * flags["clusters"] + [(None, None)] * features_dim * flags[  # type: ignore[operator]
        "clusters"
    ]
    LFR_optim_objective.steps = 0  # type: ignore[attr-defined]

    learned_model = optim.fmin_l_bfgs_b(  # type: ignore[attr-defined]
        LFR_optim_objective,
        x0=parameters_initialization,
        epsilon=1e-5,
        args=(
            training_nonsensitive,
            training_sensitive,
            ytrain_nonsensitive,
            ytrain_sensitive,
            flags["clusters"],
            flags["Ax"],
            flags["Ay"],
            flags["Az"],
            print_interval,
            verbose,
        ),
        bounds=bnd,
        approx_grad=True,
        maxfun=flags["maxfun"],
        maxiter=flags["max_iter"],
        disp=verbose,
    )[0]
    w = learned_model[: flags["clusters"]]
    prototypes = learned_model[flags["clusters"] :].reshape((flags["clusters"], features_dim))

    return Model(prototypes=prototypes, w=w)


def trans(
    prototypes: np.ndarray, w: np.ndarray, nonsens: np.ndarray, sens: np.ndarray, dataset: TestTuple
) -> pd.DataFrame:
    """Trans.

    :param prototypes:
    :param w:
    :param nonsens:
    :param sens:
    :param dataset:
    """
    _, features_hat_nonsensitive, _ = get_xhat_y_hat(prototypes, w, nonsens)

    _, features_hat_sensitive, _ = get_xhat_y_hat(prototypes, w, sens)

    sensitive_idx = dataset.x[dataset.s == 0].index
    nonsensitive_idx = dataset.x[dataset.s == 1].index

    transformed_features = np.zeros_like(dataset.x.to_numpy())
    transformed_features[sensitive_idx] = features_hat_sensitive
    transformed_features[nonsensitive_idx] = features_hat_nonsensitive

    return pd.DataFrame(transformed_features, columns=dataset.x.columns)


def main() -> None:
    """LFR Model.

    Learning fair representations is a pre-processing technique that finds a
    latent representation which encodes the data well but obfuscates information
    about protected attributes [2]_.

    References:
        .. [2] R. Zemel, Y. Wu, K. Swersky, T. Pitassi, and C. Dwork,  "Learning
           Fair Representations." International Conference on Machine Learning,
           2013.

    Based on code from https://github.com/zjelveh/learning-fair-representations
    Which in turn, we've got from AIF360
    """
    pre_algo_args: PreAlgoArgs = json.loads(sys.argv[1])
    flags: ZemelArgs = json.loads(sys.argv[2])
    if pre_algo_args["mode"] == "run":
        train, test = load_data_from_flags(pre_algo_args)
        save_transformations(
            train_and_transform(train, test, flags, seed=pre_algo_args["seed"]), pre_algo_args
        )
    elif pre_algo_args["mode"] == "fit":
        train = DataTuple.from_npz(Path(pre_algo_args["train"]))
        model = fit(train, flags, seed=pre_algo_args["seed"])
        training_sensitive = train.x.loc[train.s == 0].to_numpy()
        training_nonsensitive = train.x.loc[train.s == 1].to_numpy()
        train_transformed = trans(
            model.prototypes, model.w, training_nonsensitive, training_sensitive, train
        )
        data = train.replace(x=train_transformed)
        data.to_npz(Path(pre_algo_args["new_train"]))
        dump(model, Path(pre_algo_args["model"]))
    elif pre_algo_args["mode"] == "transform":
        model = load(Path(pre_algo_args["model"]))
        transformed_test = transform(
            DataTuple.from_npz(Path(pre_algo_args["test"])), model.prototypes, model.w
        )
        transformed_test.to_npz(Path(pre_algo_args["new_test"]))


if __name__ == "__main__":
    main()



"""Kamiran&Calders 2012, massaging."""
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

import pandas as pd
from ranzen import implements

from ethicml.utility import DataTuple, SoftPrediction, concat

from ..inprocess.logistic_regression import LR
from .pre_algorithm import PreAlgorithm, T

__all__ = ["Calders"]


@dataclass
class Calders(PreAlgorithm):
    """Massaging algorithm from Kamiran&Calders 2012."""

    preferable_class: int = 1
    disadvantaged_group: int = 0

    def __post_init__(self) -> None:
        self._out_size: Optional[int] = None

    @implements(PreAlgorithm)
    def get_out_size(self) -> int:
        assert self._out_size is not None
        return self._out_size

    @implements(PreAlgorithm)
    def get_name(self) -> str:
        return "Calders"

    @implements(PreAlgorithm)
    def fit(self, train: DataTuple, seed: int = 888) -> Tuple[PreAlgorithm, DataTuple]:
        self._out_size = train.x.shape[1]
        new_train, _ = _calders_algorithm(
            train, train, self.preferable_class, self.disadvantaged_group, seed
        )
        return self, new_train.rename(f"{self.name}: {train.name}")

    @implements(PreAlgorithm)
    def transform(self, data: T) -> T:
        return data.rename(f"{self.name}: {data.name}")

    @implements(PreAlgorithm)
    def run(self, train: DataTuple, test: T, seed: int = 888) -> Tuple[DataTuple, T]:
        self._out_size = train.x.shape[1]
        new_train, new_test = _calders_algorithm(
            train, test, self.preferable_class, self.disadvantaged_group, seed
        )
        return new_train.rename(name=f"{self.name}: {train.name}"), new_test.rename(
            name=f"{self.name}: {test.name}"
        )


def _calders_algorithm(
    dataset: DataTuple, test: T, good_class: int, disadvantaged_group: int, seed: int
) -> Tuple[DataTuple, T]:
    s_vals: List[int] = list(map(int, dataset.s.unique()))
    y_vals: List[int] = list(map(int, dataset.y.unique()))

    assert len(s_vals) == 2
    assert len(y_vals) == 2
    s_0, s_1 = s_vals
    y_0, y_1 = y_vals

    bad_class = y_0 if good_class == y_1 else y_1
    advantaged_group = s_0 if disadvantaged_group == s_1 else s_1

    groups = ((s_0, y_0), (s_0, y_1), (s_1, y_0), (s_1, y_1))
    data: Dict[Tuple[int, int], DataTuple] = {}
    for s, y in groups:
        s_y_mask = (dataset.s == s) & (dataset.y == y)
        data[(s, y)] = dataset.replace_data(data=dataset.data.loc[s_y_mask].reset_index(drop=True))

    dis_group = (disadvantaged_group, bad_class)
    adv_group = (advantaged_group, good_class)

    massaging_candidates = concat([data[dis_group], data[adv_group]])

    ranker = LR()
    rank: SoftPrediction = ranker.run(dataset, massaging_candidates, seed)

    dis_group_len = len(data[dis_group])
    adv_group_len = len(data[adv_group])

    dis_group_rank = pd.Series(rank.soft[:dis_group_len][:, 1])
    adv_group_rank = pd.Series(rank.soft[dis_group_len:][:, 1]).reset_index(drop=True)
    assert len(adv_group_rank) == adv_group_len

    # sort the ranking
    dis_group_rank.sort_values(ascending=False, inplace=True)
    adv_group_rank.sort_values(inplace=True)

    # use the rank to sort the data
    for group, ranking in [(dis_group, dis_group_rank), (adv_group, adv_group_rank)]:
        unsorted_data = data[group]
        data[group] = unsorted_data.replace_data(
            data=unsorted_data.data.reindex(index=ranking.index).reset_index(drop=True)
        )

    all_disadvantaged = len(data[(disadvantaged_group, good_class)]) + dis_group_len
    all_advantaged = adv_group_len + len(data[(advantaged_group, bad_class)])
    dis_group_good_len = all_disadvantaged - dis_group_len

    # ensure that the ratio of good_class to bad_class is the same in both groups.
    # for this, we have to swap some labels
    num_to_swap = round(
        (adv_group_len * all_disadvantaged - dis_group_good_len * all_advantaged) / len(dataset)
    )
    data[dis_group].y.iloc[:num_to_swap] = good_class  # type: ignore[call-overload]
    data[adv_group].y.iloc[:num_to_swap] = bad_class  # type: ignore[call-overload]

    return concat(list(data.values())), test



"""Variational Fair Auto-Encoder by Louizos et al."""
from dataclasses import dataclass, field
from typing import List
from typing_extensions import TypedDict

from ranzen import implements

from ethicml.algorithms.preprocess.pre_subprocess import PreAlgorithmSubprocess

__all__ = ["VFAE"]

from ethicml.utility import FairnessType


class VfaeArgs(TypedDict):
    """Args object of VFAE."""

    dataset: str
    supervised: bool
    epochs: int
    batch_size: int
    fairness: str
    latent_dims: int
    z1_enc_size: List[int]
    z2_enc_size: List[int]
    z1_dec_size: List[int]


@dataclass
class VFAE(PreAlgorithmSubprocess):
    """VFAE Object - see implementation file for details."""

    dataset: str = "toy"
    supervised: bool = True
    epochs: int = 10
    batch_size: int = 32
    fairness: FairnessType = FairnessType.dp
    latent_dims: int = 50
    z1_enc_size: List[int] = field(default_factory=lambda: [100])
    z2_enc_size: List[int] = field(default_factory=lambda: [100])
    z1_dec_size: List[int] = field(default_factory=lambda: [100])

    @implements(PreAlgorithmSubprocess)
    def _get_flags(self) -> VfaeArgs:
        # TODO: replace this with dataclasses.asdict()
        return {
            "supervised": self.supervised,
            "fairness": str(self.fairness),
            "batch_size": self.batch_size,
            "epochs": self.epochs,
            "dataset": self.dataset,
            "latent_dims": self.latent_dims,
            "z1_enc_size": self.z1_enc_size,
            "z2_enc_size": self.z2_enc_size,
            "z1_dec_size": self.z1_dec_size,
        }

    @implements(PreAlgorithmSubprocess)
    def get_out_size(self) -> int:
        return self.latent_dims

    @implements(PreAlgorithmSubprocess)
    def get_name(self) -> str:
        return "VFAE"

    @implements(PreAlgorithmSubprocess)
    def _get_path_to_script(self) -> List[str]:
        return ["-m", "ethicml.implementations.vfae"]



"""Wrapper for SKLearn implementation of SVM."""
from dataclasses import dataclass, field
from typing import ClassVar, Union

import numpy as np
import pandas as pd
from ranzen import implements
from sklearn.svm import SVC, LinearSVC

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmDC
from ethicml.utility import DataTuple, Prediction, TestTuple

__all__ = ["SVM"]

from ethicml.utility import KernelType


@dataclass
class SVM(InAlgorithmDC):
    """A wrapper around the SciKitLearn Support Vector Classifier (SVC) model.

    Documentation for the underlying classifier can be found
    `here <https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html>`_.

    :param C: The penalty parameter of the error term.
    :param kernel: The kernel to use.
    """

    is_fairness_algo: ClassVar[bool] = False
    C: float = field(default_factory=lambda: SVC().C)
    kernel: KernelType = field(default_factory=lambda: KernelType[SVC().kernel])

    @implements(InAlgorithmDC)
    def get_name(self) -> str:
        return f"SVM ({self.kernel})"

    @implements(InAlgorithmDC)
    def fit(self, train: DataTuple, seed: int = 888) -> "SVM":
        self.clf = select_svm(self.C, self.kernel, seed)
        self.clf.fit(train.x, train.y.to_numpy().ravel())
        return self

    @implements(InAlgorithmDC)
    def predict(self, test: TestTuple) -> Prediction:
        return Prediction(hard=pd.Series(self.clf.predict(test.x)))

    @implements(InAlgorithmDC)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        clf = select_svm(self.C, self.kernel, seed)
        clf.fit(train.x, train.y.to_numpy().ravel())
        return Prediction(hard=pd.Series(clf.predict(test.x)))


def select_svm(C: float, kernel: KernelType, seed: int) -> Union[LinearSVC, SVC]:
    """Select the appropriate SVM model for the given parameters.

    :param C: The penalty parameter of the error term.
    :param kernel: The kernel to use.
    :param seed: The seed for the random number generator.
    """
    random_state = np.random.RandomState(seed=seed)
    if kernel is KernelType.linear:
        return LinearSVC(C=C, dual=False, tol=1e-12, random_state=random_state)
    return SVC(C=C, kernel=kernel.name, gamma="auto", random_state=random_state)



"""Implementation of Beutel's adversarially learned fair representations."""
# Disable pylint checking overwritten method signatures. Pytorch forward passes use **kwargs
# pylint: disable=arguments-differ

from __future__ import annotations

import json
import random
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Optional, Sequence, Tuple

import numpy as np
import pandas as pd
import torch
from joblib import dump, load
from torch import Tensor, nn
from torch.autograd import Function
from torch.optim import Adam
from torch.optim.lr_scheduler import ExponentialLR

from ethicml.preprocessing.adjust_labels import LabelBinarizer, assert_binary_labels
from ethicml.preprocessing.splits import train_test_split
from ethicml.utility import DataTuple, FairnessType

from .pytorch_common import CustomDataset, TestDataset, make_dataset_and_loader
from .utils import load_data_from_flags, save_transformations

if TYPE_CHECKING:
    from ethicml.algorithms.preprocess.beutel import BeutelArgs
    from ethicml.algorithms.preprocess.pre_subprocess import PreAlgoArgs, T

STRING_TO_ACTIVATION_MAP = {"Sigmoid()": nn.Sigmoid()}

STRING_TO_LOSS_MAP = {"BCELoss()": nn.BCELoss(), "CrossEntropyLoss()": nn.CrossEntropyLoss()}


def set_seed(seed: int) -> None:
    """Set the seeds for numpy torch etc.

    :param seed:
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def build_networks(
    flags: BeutelArgs,
    train_data: CustomDataset,
    enc_activation: nn.Module,
    adv_activation: nn.Module,
) -> Tuple[Encoder, Model]:
    """Build the networks we use.

    Pulled into a separate function to make the code a bit neater.

    :param flags:
    :param train_data:
    :param enc_activation:
    :param adv_activation:
    """
    enc = Encoder(
        enc_size=flags["enc_size"], init_size=int(train_data.xdim), activation=enc_activation
    )
    adv = Adversary(
        fairness=FairnessType[flags["fairness"]],
        adv_size=flags["adv_size"],
        init_size=flags["enc_size"][-1],
        s_size=int(train_data.sdim),
        activation=adv_activation,
        adv_weight=flags["adv_weight"],
    )
    pred = Predictor(
        pred_size=flags["pred_size"],
        init_size=flags["enc_size"][-1],
        class_label_size=len(np.unique(train_data.y)),
        activation=adv_activation,
    )

    model = Model(enc, adv, pred)

    return enc, model


def fit(train: DataTuple, flags: BeutelArgs, seed: int = 888) -> Tuple[DataTuple, Encoder]:
    """Train the fair autoencoder on the training data and then transform both training and test.

    :param train:
    :param flags:
    """
    set_seed(seed)
    fairness = FairnessType[flags["fairness"]]

    post_process = False
    if flags["y_loss"] == "BCELoss()":
        try:
            assert_binary_labels(train)
        except AssertionError:
            processor = LabelBinarizer()
            train = processor.adjust(train)
            post_process = True

    # By default we use 10% of the training data for validation
    train_, validation = train_test_split(train, train_percentage=1 - flags["validation_pcnt"])

    train_data, train_loader = make_dataset_and_loader(
        train_, batch_size=flags["batch_size"], shuffle=True, seed=seed, drop_last=True
    )
    _, validation_loader = make_dataset_and_loader(
        validation, batch_size=flags["batch_size"], shuffle=False, seed=seed, drop_last=True
    )
    _, all_train_data_loader = make_dataset_and_loader(
        train, batch_size=flags["batch_size"], shuffle=False, seed=seed, drop_last=True
    )

    # convert flags to Python objects
    enc_activation = STRING_TO_ACTIVATION_MAP[flags["enc_activation"]]
    adv_activation = STRING_TO_ACTIVATION_MAP[flags["adv_activation"]]
    y_loss_fn = STRING_TO_LOSS_MAP[flags["y_loss"]]
    s_loss_fn = STRING_TO_LOSS_MAP[flags["s_loss"]]

    enc, model = build_networks(
        flags=flags,
        train_data=train_data,
        enc_activation=enc_activation,
        adv_activation=adv_activation,
    )

    optimizer = torch.optim.Adam(model.parameters())
    scheduler = ExponentialLR(optimizer, gamma=0.95)

    best_val_loss = torch.ones(1) * np.inf
    best_enc = None

    for i in range(1, flags["epochs"] + 1):
        model.train()
        for embedding, sens_label, class_label in train_loader:
            sens_label = sens_label.view(-1, 1)
            class_label = class_label.view(-1, 1)
            _, s_pred, y_pred = model(embedding, class_label)

            loss = y_loss_fn(y_pred, class_label.squeeze(-1).long())

            if fairness is FairnessType.eq_opp:
                mask = class_label.ge(0.5)
            elif fairness is FairnessType.eq_odds:
                raise NotImplementedError("Not implemented Eq. Odds yet")
            elif fairness is FairnessType.dp:
                mask = torch.ones(s_pred.shape, dtype=torch.uint8)
            loss += s_loss_fn(
                s_pred, torch.masked_select(sens_label, mask).view(-1, int(train_data.sdim))
            )

            step(i, loss, optimizer, scheduler)

        if i % 5 == 0 or i == flags["epochs"]:
            model.eval()
            val_y_loss = torch.zeros(1)
            val_s_loss = torch.zeros(1)
            for embedding, sens_label, class_label in validation_loader:
                sens_label = sens_label.view(-1, 1)
                class_label = class_label.view(-1, 1)
                _, s_pred, y_pred = model(embedding, class_label)

                val_y_loss += y_loss_fn(y_pred, class_label.squeeze(-1).long())

                mask = get_mask(flags, s_pred, class_label)

                val_s_loss -= s_loss_fn(
                    s_pred, torch.masked_select(sens_label, mask).view(-1, int(train_data.sdim))
                )

            val_loss = (val_y_loss / len(validation_loader)) + (val_s_loss / len(validation_loader))

            if val_loss < best_val_loss:
                best_val_loss = val_loss
                best_enc = enc.state_dict()

    assert best_enc is not None
    enc.load_state_dict(best_enc)

    transformed_train = encode_dataset(enc, all_train_data_loader, train)
    if post_process:
        transformed_train = processor.post(encode_dataset(enc, all_train_data_loader, train))
    return transformed_train, enc


def transform(data: T, enc: torch.nn.Module, flags: BeutelArgs) -> T:
    """Transform the test data using the trained autoencoder.

    :param data:
    :param enc:
    :param flags:
    """
    test_data = TestDataset(data)
    test_loader = torch.utils.data.DataLoader(
        dataset=test_data, batch_size=flags["batch_size"], shuffle=False
    )
    return encode_testset(enc, test_loader, data)


def train_and_transform(
    train: DataTuple, test: T, flags: BeutelArgs, seed: int
) -> Tuple[DataTuple, T]:
    """Train the fair autoencoder on the training data and then transform both training and test.

    :param train:
    :param test:
    :param flags:
    """
    transformed_train, enc = fit(train, flags, seed)
    transformed_test = transform(test, enc, flags)
    return transformed_train, transformed_test


def step(iteration: int, loss: Tensor, optimizer: Adam, scheduler: ExponentialLR) -> None:
    """Do one training step.

    :param iteration:
    :param loss:
    :param optimizer:
    :param scheduler:
    """
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()
    scheduler.step(iteration)


def get_mask(flags: BeutelArgs, s_pred: Tensor, class_label: Tensor) -> Tensor:
    """Get a mask to enforce different fairness types.

    :param flags:
    :param s_pred:
    :param class_label:
    """
    fairness = FairnessType[flags["fairness"]]
    if fairness is FairnessType.eq_opp:
        mask = class_label.ge(0.5)
    elif fairness is FairnessType.eq_odds:
        raise NotImplementedError("Not implemented Eq. Odds yet")
    elif fairness is FairnessType.dp:
        mask = torch.ones(s_pred.shape, dtype=torch.uint8)
    return mask


def encode_dataset(
    enc: nn.Module, dataloader: torch.utils.data.DataLoader, datatuple: DataTuple
) -> DataTuple:
    """Encode a dataset.

    :param enc:
    :param dataloader:
    :param datatuple:
    """
    data_to_return: List[Any] = []

    for embedding, _, _ in dataloader:
        data_to_return += enc(embedding).data.numpy().tolist()

    return datatuple.replace(x=pd.DataFrame(data_to_return)).rename(f"Beutel: {datatuple.name}")


def encode_testset(enc: nn.Module, dataloader: torch.utils.data.DataLoader, testtuple: T) -> T:
    """Encode a dataset.

    :param enc:
    :param dataloader:
    :param testtuple:
    :returns: Encoded TestTuple.
    """
    data_to_return: List[Any] = []

    for embedding, _ in dataloader:
        data_to_return += enc(embedding).data.numpy().tolist()

    return testtuple.replace(x=pd.DataFrame(data_to_return))


class GradReverse(Function):
    """Gradient reversal layer."""

    @staticmethod
    def forward(ctx: Any, x: Tensor, lambda_: float) -> Any:  # type: ignore[override]
        """Forward pass.

        :param ctx:
        :param x:
        :param lambda_:
        """
        ctx.lambda_ = lambda_
        return x.view_as(x)

    @staticmethod
    def backward(ctx: Any, grad_output: Tensor) -> Any:  # type: ignore[override]
        """Backward pass with Gradient reversed / inverted.

        :param ctx:
        :param grad_output:
        """
        return grad_output.neg().mul(ctx.lambda_), None


def _grad_reverse(features: Tensor, lambda_: float) -> Tensor:
    return GradReverse.apply(features, lambda_)


class Encoder(nn.Module):
    """Encoder of the GAN."""

    def __init__(self, enc_size: Sequence[int], init_size: int, activation: Optional[nn.Module]):
        super().__init__()
        self.encoder = nn.Sequential()
        if not enc_size:  # In the case that encoder size [] is specified
            self.encoder.add_module("single encoder layer", nn.Linear(init_size, init_size))
            self.encoder.add_module("single layer encoder activation", activation)
        else:
            self.encoder.add_module("encoder layer 0", nn.Linear(init_size, enc_size[0]))
            self.encoder.add_module("encoder activation 0", activation)
            for k in range(len(enc_size) - 1):
                self.encoder.add_module(
                    f"encoder layer {k + 1}", nn.Linear(enc_size[k], enc_size[k + 1])
                )
                self.encoder.add_module(f"encoder activation {k + 1}", activation)

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass.

        :param x:
        """
        return self.encoder(x)


class Adversary(nn.Module):
    """Adversary of the GAN."""

    def __init__(
        self,
        fairness: FairnessType,
        adv_size: Sequence[int],
        init_size: int,
        s_size: int,
        activation: nn.Module,
        adv_weight: float,
    ):
        super().__init__()
        self.fairness = fairness
        self.init_size = init_size
        self.adv_weight = adv_weight
        self.adversary = nn.Sequential()
        if not adv_size:  # In the case that encoder size [] is specified
            self.adversary.add_module("single adversary layer", nn.Linear(init_size, s_size))
            self.adversary.add_module("single layer adversary activation", activation)
        else:
            self.adversary.add_module("adversary layer 0", nn.Linear(init_size, adv_size[0]))
            self.adversary.add_module("adversary activation 0", activation)
            for k in range(len(adv_size) - 1):
                self.adversary.add_module(
                    f"adversary layer {k + 1}", nn.Linear(adv_size[k], adv_size[k + 1])
                )
                self.adversary.add_module(f"adversary activation {k + 1}", activation)
            self.adversary.add_module("adversary last layer", nn.Linear(adv_size[-1], s_size))
            self.adversary.add_module("adversary last activation", activation)

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        """Forward pass.

        :param x:
        :param y:
        """
        x = _grad_reverse(x, lambda_=self.adv_weight)

        if self.fairness is FairnessType.eq_opp:
            mask = y.view(-1, 1).ge(0.5)
            x = torch.masked_select(x, mask).view(-1, self.init_size)
            x = self.adversary(x)
        elif self.fairness is FairnessType.eq_odds:
            raise NotImplementedError("Not implemented equalized odds yet")
        elif self.fairness is FairnessType.dp:
            x = self.adversary(x)
        else:
            raise NotImplementedError("Shouldn't be hit.")
        return x


class Predictor(nn.Module):
    """Predictor of the GAN."""

    def __init__(
        self, pred_size: Sequence[int], init_size: int, class_label_size: int, activation: nn.Module
    ):
        super().__init__()
        self.predictor = nn.Sequential()
        if not pred_size:  # In the case that encoder size [] is specified
            self.predictor.add_module(
                "single adversary layer", nn.Linear(init_size, class_label_size)
            )
            self.predictor.add_module("single layer adversary activation", activation)
        else:
            self.predictor.add_module("adversary layer 0", nn.Linear(init_size, pred_size[0]))
            self.predictor.add_module("adversary activation 0", nn.Sigmoid())
            for k in range(len(pred_size) - 1):
                self.predictor.add_module(
                    f"adversary layer {k + 1}", nn.Linear(pred_size[k], pred_size[k + 1])
                )
                self.predictor.add_module(f"adversary activation {k + 1}", nn.Sigmoid())
            self.predictor.add_module(
                "adversary last layer", nn.Linear(pred_size[-1], class_label_size)
            )
            self.predictor.add_module("adversary last activation", nn.Softmax())

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass.

        :param x:
        """
        return self.predictor(x)


class Model(nn.Module):
    """Whole GAN model."""

    def __init__(self, enc: Encoder, adv: Adversary, pred: Predictor) -> None:
        super().__init__()
        self.enc = enc
        self.adv = adv
        self.pred = pred

    def forward(self, x: Tensor, y: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """Forward pass.

        :param x:
        :param y:
        """
        encoded = self.enc(x)
        s_hat = self.adv(encoded, y)
        y_hat = self.pred(encoded)
        return encoded, s_hat, y_hat


def main() -> None:
    """Load data from feather files, pass it to `train_and_transform` and then save the result."""
    pre_algo_args: PreAlgoArgs = json.loads(sys.argv[1])
    flags: BeutelArgs = json.loads(sys.argv[2])
    if pre_algo_args["mode"] == "run":
        train, test = load_data_from_flags(pre_algo_args)
        save_transformations(
            train_and_transform(train, test, flags, pre_algo_args["seed"]), pre_algo_args
        )
    elif pre_algo_args["mode"] == "fit":
        train = DataTuple.from_npz(Path(pre_algo_args["train"]))
        transformed_train, enc = fit(train, flags, seed=pre_algo_args["seed"])
        transformed_train.to_npz(Path(pre_algo_args["new_train"]))
        dump(enc, Path(pre_algo_args["model"]))
    elif pre_algo_args["mode"] == "transform":
        model = load(Path(pre_algo_args["model"]))
        transformed_test = transform(DataTuple.from_npz(Path(pre_algo_args["test"])), model, flags)
        transformed_test.to_npz(Path(pre_algo_args["new_test"]))


if __name__ == "__main__":
    main()



"""Scale a dataset."""
from typing import Tuple
from typing_extensions import Protocol

import pandas as pd

from ethicml.data.dataset import Dataset
from ethicml.utility import DataTuple

__all__ = ["scale_continuous"]


class ScalerType(Protocol):
    def fit(self, df: pd.DataFrame) -> None:
        ...

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        ...

    def inverse_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        ...

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        ...


def scale_continuous(
    dataset: Dataset,
    datatuple: DataTuple,
    scaler: ScalerType,
    inverse: bool = False,
    fit: bool = True,
) -> Tuple[DataTuple, ScalerType]:
    """Use a scaler on just the continuous features.

    :param dataset: Dataset object. Used to find the continuous features.
    :param datatuple: DataTuple on which to sclae the continuous features.
    :param scaler: Scaler object to scale the features. Must fit the SKLearn scaler API.
    :param inverse: Should the scaling be reversed? (Default: False)
    :param fit: If not `inverse`, should the scaler be fit to the data? If `True`, do
        `fit_transform` operation, else just `transform`. (Default: True)
    :returns: Tuple of (scaled) DataTuple, and the Scaler (which may have been fit to the data).

    Examples:
        >>> dataset = adult()
        >>> datatuple = dataset.load()
        >>> train, test = train_test_split(datatuple)
        >>> train, scaler = scale_continuous(dataset, train, scaler)
        >>> test, scaler = scale_continuous(dataset, test, scaler, fit=False)
    """
    new_feats = datatuple.x.copy().astype('float64')
    if inverse:
        new_feats[dataset.continuous_features] = scaler.inverse_transform(
            new_feats[dataset.continuous_features]
        )
    elif fit:
        new_feats[dataset.continuous_features] = scaler.fit_transform(
            new_feats[dataset.continuous_features]
        )
    else:
        new_feats[dataset.continuous_features] = scaler.transform(
            new_feats[dataset.continuous_features]
        )
    return datatuple.replace(x=new_feats), scaler



"""Classes related to running algorithms in subprocesses."""
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, List, Mapping, Optional, Tuple, TypeVar, Union
from typing_extensions import Literal, TypeAlias, TypedDict, final

from ethicml.algorithms.algorithm_base import SubprocessAlgorithmMixin
from ethicml.algorithms.preprocess.pre_algorithm import PreAlgorithm
from ethicml.utility import DataTuple, SubgroupTuple

__all__ = ["PreAlgoArgs", "PreAlgorithmSubprocess"]

T = TypeVar("T", DataTuple, SubgroupTuple)


class PreAlgoRunArgs(TypedDict):
    """Base arguments for the ``run`` function of async pre-process methods."""

    mode: Literal["run"]
    train: str
    test: str
    # paths to where the processed inputs should be stored
    new_train: str
    new_test: str
    seed: int


class PreAlgoFitArgs(TypedDict):
    """Base arguments for the ``fit`` function of async pre-process methods."""

    mode: Literal["fit"]
    train: str
    new_train: str
    model: str  # path to where the model weights are stored
    seed: int


class PreAlgoTformArgs(TypedDict):
    """Base arguments for the ``transform`` function of async pre-process methods."""

    mode: Literal["transform"]
    test: str
    new_test: str
    model: str


PreAlgoArgs: TypeAlias = Union[PreAlgoFitArgs, PreAlgoTformArgs, PreAlgoRunArgs]
_P = TypeVar("_P", bound="PreAlgorithmSubprocess")


@dataclass  # type: ignore[misc]  # mypy doesn't allow abstract dataclasses because mypy is stupid
class PreAlgorithmSubprocess(SubprocessAlgorithmMixin, PreAlgorithm, ABC):
    """Pre-Algorithm that runs the method in a subprocess.

    This is the base class for all pre-processing algorithms that run in a subprocess. The
    advantage of this is that it allows for parallelization.
    """

    dir: Path = Path(".")

    @property
    @final
    def model_path(self) -> Path:
        """Path to where the model with be stored."""
        return self.dir.resolve(strict=True) / f"model_{self.name}.joblib"

    @final
    def fit(self: _P, train: DataTuple, seed: int = 888) -> Tuple[_P, DataTuple]:
        """Fit transformer in a subprocess on the given data.

        :param train: Data tuple of the training data.
        :param seed: Random seed for model initialization.
        :returns: A tuple of Self and the test data.
        """
        self._in_size: Optional[int] = train.x.shape[1]
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            # ================================ write data to files ================================
            train_path = tmp_path / "train.npz"
            train.to_npz(train_path)

            # ========================== generate commandline arguments ===========================
            transformed_train_path = tmp_path / "transformed_train.npz"
            args: PreAlgoFitArgs = {
                "mode": "fit",
                "model": str(self.model_path),
                "train": str(train_path),
                "new_train": str(transformed_train_path),
                "seed": seed,
            }

            # ============================= run the generated command =============================
            self.call_script(self._script_command(args))

            # ================================== load results =====================================
            transformed_train = DataTuple.from_npz(transformed_train_path)

        # prefix the name of the algorithm to the dataset name
        if train.name is not None:
            transformed_train = transformed_train.rename(f"{self.name}: {train.name}")
        return self, transformed_train

    @final
    def transform(self, data: T) -> T:
        """Generate fair features in a subprocess with the given data.

        :param data: Data to transform.
        :returns: Transformed data.
        """
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            # ================================ write data to files ================================
            test_path = tmp_path / "test.npz"
            data.to_npz(test_path)

            # ========================== generate commandline arguments ===========================
            transformed_test_path = tmp_path / "transformed_test.npz"
            args: PreAlgoTformArgs = {
                "mode": "transform",
                "model": str(self.model_path),
                "test": str(test_path),
                "new_test": str(transformed_test_path),
            }

            # ============================= run the generated command =============================
            self.call_script(self._script_command(args))

            # ================================== load results =====================================
            transformed_test: T = data.from_npz(transformed_test_path)

        # prefix the name of the algorithm to the dataset name
        if data.name is not None:
            transformed_test = transformed_test.rename(f"{self.name}: {data.name}")
        return transformed_test

    @final
    def run(self, train: DataTuple, test: T, seed: int = 888) -> Tuple[DataTuple, T]:
        """Generate fair features in a subprocess with the given data.

        :param train: Data tuple of the training data.
        :param test: Data tuple of the test data.
        :param seed: Random seed for model initialization.
        :returns: A tuple of the transforme training data and the test data.
        """
        self._in_size = train.x.shape[1]
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            # ================================ write data to files ================================
            train_path, test_path = tmp_path / "train.npz", tmp_path / "test.npz"
            train.to_npz(train_path)
            test.to_npz(test_path)

            # ========================== generate commandline arguments ===========================
            transformed_train_path = tmp_path / "transformed_train.npz"
            transformed_test_path = tmp_path / "transformed_test.npz"
            args: PreAlgoRunArgs = {
                "mode": "run",
                "train": str(train_path),
                "test": str(test_path),
                "new_train": str(transformed_train_path),
                "new_test": str(transformed_test_path),
                "seed": seed,
            }

            # ============================= run the generated command =============================
            self.call_script(self._script_command(args))

            # ================================== load results =====================================
            transformed_train = DataTuple.from_npz(transformed_train_path)
            transformed_test = test.from_npz(transformed_test_path)

        # prefix the name of the algorithm to the dataset name
        if train.name is not None:
            transformed_train = transformed_train.rename(f"{self.name}: {train.name}")
        if test.name is not None:
            transformed_test = transformed_test.rename(f"{self.name}: {test.name}")
        return transformed_train, transformed_test

    @final
    def _script_command(self, pre_algo_args: PreAlgoArgs) -> List[str]:
        """Return the command that will run the script.

        The flag interface consists of two strings, both JSON strings: the general pre-algo flags
        and then the more specific flags for the algorithm.

        :param pre_algo_args: The Arguments that are shared among all pre-processing algorithms.
        :return: List of strings that can be passed to ``subprocess.run``.
        """
        interface = [
            json.dumps(pre_algo_args, separators=(',', ':')),
            json.dumps(self._get_flags(), separators=(',', ':')),
        ]
        return self._get_path_to_script() + interface

    @abstractmethod
    def _get_path_to_script(self) -> List[str]:
        """Return arguments that are passed to the python executable."""

    @abstractmethod
    def _get_flags(self) -> Mapping[str, Any]:
        """Return flags that are used to configure this algorithm."""



"""Load Data from .csv files."""
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

import pandas as pd

from ethicml.utility import DataTuple

from .dataset import Dataset, LoadableDataset

__all__ = ["load_data", "create_data_obj"]


def load_data(dataset: Dataset, ordered: bool = False) -> DataTuple:
    """Load dataset from its CSV file.

    This function only exists for backwards compatibility. Use dataset.load() instead.

    :param dataset: dataset object
    :param ordered: if True, return features such that discrete come first, then continuous (Default: False)
    :returns: DataTuple with dataframes of features, labels and sensitive attributes
    """
    return dataset.load(ordered=ordered)


def create_data_obj(
    filepath: Path, s_column: str, y_column: str, additional_to_drop: Optional[List[str]] = None
) -> "ConfigurableDataset":
    """Create a `ConfigurableDataset` from the given file.

    :param filepath: path to a CSV file
    :param s_column: column that represents sensitive attributes
    :param y_column: column that contains lables
    :param additional_to_drop: other columns that should be dropped (Default: None)
    :returns: Dataset object
    """
    return ConfigurableDataset(
        filepath_=filepath,
        s_column=s_column,
        y_column=y_column,
        additional_to_drop=additional_to_drop,
    )


@dataclass
class ConfigurableDataset(LoadableDataset):
    """A configurable dataset class."""

    filepath_: Optional[Path] = None
    s_column: Optional[str] = None
    y_column: Optional[str] = None
    additional_to_drop: Optional[List[str]] = None

    def __post_init__(self) -> None:
        assert self.filepath_ is not None
        assert self.s_column is not None
        assert self.y_column is not None
        if self.additional_to_drop is None:
            self.additional_to_drop = []

        dataframe: pd.DataFrame = pd.read_csv(self.filepath_)

        columns: List[str] = [str(x) for x in dataframe.columns.to_numpy().tolist()]
        columns.remove(self.s_column)
        columns.remove(self.y_column)
        for additional in self.additional_to_drop:
            columns.remove(additional)

        super().__init__(
            name=self.filepath_.name,
            num_samples=len(dataframe),
            features=columns,
            cont_features=[],
            sens_attr_spec=self.s_column,
            class_label_spec=self.y_column,
            filename_or_path=self.filepath_,
            discrete_only=False,
        )



"""Class to describe features of the Compas dataset."""
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type, Union

from ..dataset import LoadableDataset
from ..util import LabelSpec, flatten_dict, simple_spec

__all__ = ["Compas", "CompasSplits"]


class CompasSplits(Enum):
    """Available dataset splits for the COMPAS dataset."""

    SEX = "Sex"
    RACE = "Race"
    RACE_SEX = "Race-Sex"
    CUSTOM = "Custom"


@dataclass
class Compas(LoadableDataset):
    """Compas (or ProPublica) dataset."""

    split: CompasSplits = CompasSplits.SEX

    Splits: ClassVar[Type[CompasSplits]] = CompasSplits
    """Shorthand for the Enum that defines the splits associated with this class."""

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "sex": ["sex"],
            "race": ["race"],
            "two-year-recid": ["two-year-recid"],
            "age-cat": ["age-cat_25 - 45", "age-cat_Greater than 45", "age-cat_Less than 25"],
            "c-charge-degree": ["c-charge-degree_F", "c-charge-degree_M"],
            "c-charge-desc": [
                "c-charge-desc_Abuse Without Great Harm",
                "c-charge-desc_Agg Abuse Elderlly/Disabled Adult",
                "c-charge-desc_Agg Assault W/int Com Fel Dome",
                "c-charge-desc_Agg Battery Grt/Bod/Harm",
                "c-charge-desc_Agg Fleeing and Eluding",
                "c-charge-desc_Agg Fleeing/Eluding High Speed",
                "c-charge-desc_Aggr Child Abuse-Torture,Punish",
                "c-charge-desc_Aggrav Battery w/Deadly Weapon",
                "c-charge-desc_Aggrav Child Abuse-Agg Battery",
                "c-charge-desc_Aggrav Child Abuse-Causes Harm",
                "c-charge-desc_Aggrav Stalking After Injunctn",
                "c-charge-desc_Aggravated Assault",
                "c-charge-desc_Aggravated Assault W/Dead Weap",
                "c-charge-desc_Aggravated Assault W/dead Weap",
                "c-charge-desc_Aggravated Assault W/o Firearm",
                "c-charge-desc_Aggravated Assault w/Firearm",
                "c-charge-desc_Aggravated Battery",
                "c-charge-desc_Aggravated Battery (Firearm)",
                "c-charge-desc_Aggravated Battery (Firearm/Actual Possession)",
                "c-charge-desc_Aggravated Battery / Pregnant",
                "c-charge-desc_Aggravated Battery On 65/Older",
                "c-charge-desc_Aide/Abet Prostitution Lewdness",
                "c-charge-desc_Aiding Escape",
                "c-charge-desc_Alcoholic Beverage Violation-FL",
                "c-charge-desc_Armed Trafficking in Cannabis",
                "c-charge-desc_Arson in the First Degree",
                "c-charge-desc_Assault",
                "c-charge-desc_Assault Law Enforcement Officer",
                "c-charge-desc_Att Burgl Conv Occp",
                "c-charge-desc_Att Burgl Struc/Conv Dwel/Occp",
                "c-charge-desc_Att Burgl Unoccupied Dwel",
                "c-charge-desc_Att Tamper w/Physical Evidence",
                "c-charge-desc_Attempt Armed Burglary Dwell",
                "c-charge-desc_Attempted Burg/Convey/Unocc",
                "c-charge-desc_Attempted Burg/struct/unocc",
                "c-charge-desc_Attempted Deliv Control Subst",
                "c-charge-desc_Attempted Robbery  No Weapon",
                "c-charge-desc_Attempted Robbery  Weapon",
                "c-charge-desc_Battery",
                "c-charge-desc_Battery Emergency Care Provide",
                "c-charge-desc_Battery On A Person Over 65",
                "c-charge-desc_Battery On Fire Fighter",
                "c-charge-desc_Battery On Parking Enfor Speci",
                "c-charge-desc_Battery Spouse Or Girlfriend",
                "c-charge-desc_Battery on Law Enforc Officer",
                "c-charge-desc_Battery on a Person Over 65",
                "c-charge-desc_Bribery Athletic Contests",
                "c-charge-desc_Burgl Dwel/Struct/Convey Armed",
                "c-charge-desc_Burglary Assault/Battery Armed",
                "c-charge-desc_Burglary Conveyance Armed",
                "c-charge-desc_Burglary Conveyance Assault/Bat",
                "c-charge-desc_Burglary Conveyance Occupied",
                "c-charge-desc_Burglary Conveyance Unoccup",
                "c-charge-desc_Burglary Dwelling Armed",
                "c-charge-desc_Burglary Dwelling Assault/Batt",
                "c-charge-desc_Burglary Dwelling Occupied",
                "c-charge-desc_Burglary Structure Assault/Batt",
                "c-charge-desc_Burglary Structure Occupied",
                "c-charge-desc_Burglary Structure Unoccup",
                "c-charge-desc_Burglary Unoccupied Dwelling",
                "c-charge-desc_Burglary With Assault/battery",
                "c-charge-desc_Carjacking w/o Deadly Weapon",
                "c-charge-desc_Carjacking with a Firearm",
                "c-charge-desc_Carry Open/Uncov Bev In Pub",
                "c-charge-desc_Carrying A Concealed Weapon",
                "c-charge-desc_Carrying Concealed Firearm",
                "c-charge-desc_Cash Item w/Intent to Defraud",
                "c-charge-desc_Child Abuse",
                "c-charge-desc_Computer Pornography",
                "c-charge-desc_Consp Traff Oxycodone  4g><14g",
                "c-charge-desc_Conspiracy Dealing Stolen Prop",
                "c-charge-desc_Consume Alcoholic Bev Pub",
                "c-charge-desc_Contradict Statement",
                "c-charge-desc_Contribute Delinquency Of A Minor",
                "c-charge-desc_Corrupt Public Servant",
                "c-charge-desc_Counterfeit Lic Plates/Sticker",
                "c-charge-desc_Crim Attempt/Solic/Consp",
                "c-charge-desc_Crim Use of Personal ID Info",
                "c-charge-desc_Crimin Mischief Damage $1000+",
                "c-charge-desc_Criminal Mischief",
                "c-charge-desc_Criminal Mischief Damage <$200",
                "c-charge-desc_Criminal Mischief>$200<$1000",
                "c-charge-desc_Crlty Twrd Child Urge Oth Act",
                "c-charge-desc_Cruelty Toward Child",
                "c-charge-desc_Cruelty to Animals",
                "c-charge-desc_Culpable Negligence",
                "c-charge-desc_D.U.I. Serious Bodily Injury",
                "c-charge-desc_DOC/Cause Public Danger",
                "c-charge-desc_DUI - Enhanced",
                "c-charge-desc_DUI - Property Damage/Personal Injury",
                "c-charge-desc_DUI Blood Alcohol Above 0.20",
                "c-charge-desc_DUI Level 0.15 Or Minor In Veh",
                "c-charge-desc_DUI Property Damage/Injury",
                "c-charge-desc_DUI- Enhanced",
                "c-charge-desc_DUI/Property Damage/Persnl Inj",
                "c-charge-desc_DWI w/Inj Susp Lic / Habit Off",
                "c-charge-desc_DWLS Canceled Disqul 1st Off",
                "c-charge-desc_DWLS Susp/Cancel Revoked",
                "c-charge-desc_Dealing in Stolen Property",
                "c-charge-desc_Defrauding Innkeeper",
                "c-charge-desc_Defrauding Innkeeper $300/More",
                "c-charge-desc_Del 3,4 Methylenedioxymethcath",
                "c-charge-desc_Del Cannabis At/Near Park",
                "c-charge-desc_Del Cannabis For Consideration",
                "c-charge-desc_Del of JWH-250 2-Methox 1-Pentyl",
                "c-charge-desc_Deliver 3,4 Methylenediox",
                "c-charge-desc_Deliver Alprazolam",
                "c-charge-desc_Deliver Cannabis",
                "c-charge-desc_Deliver Cannabis 1000FTSch",
                "c-charge-desc_Deliver Cocaine",
                "c-charge-desc_Deliver Cocaine 1000FT Church",
                "c-charge-desc_Deliver Cocaine 1000FT Park",
                "c-charge-desc_Deliver Cocaine 1000FT School",
                "c-charge-desc_Deliver Cocaine 1000FT Store",
                "c-charge-desc_Delivery Of Drug Paraphernalia",
                "c-charge-desc_Delivery of 5-Fluoro PB-22",
                "c-charge-desc_Delivery of Heroin",
                "c-charge-desc_Depriv LEO of Protect/Communic",
                "c-charge-desc_Disorderly Conduct",
                "c-charge-desc_Disorderly Intoxication",
                "c-charge-desc_Disrupting School Function",
                "c-charge-desc_Drivg While Lic Suspd/Revk/Can",
                "c-charge-desc_Driving License Suspended",
                "c-charge-desc_Driving Under The Influence",
                "c-charge-desc_Driving While License Revoked",
                "c-charge-desc_Escape",
                "c-charge-desc_Exhibition Weapon School Prop",
                "c-charge-desc_Expired DL More Than 6 Months",
                "c-charge-desc_Exposes Culpable Negligence",
                "c-charge-desc_Extradition/Defendants",
                "c-charge-desc_Fabricating Physical Evidence",
                "c-charge-desc_Fail Register Vehicle",
                "c-charge-desc_Fail Sex Offend Report Bylaw",
                "c-charge-desc_Fail To Obey Police Officer",
                "c-charge-desc_Fail To Redeliv Hire/Leas Prop",
                "c-charge-desc_Failure To Pay Taxi Cab Charge",
                "c-charge-desc_Failure To Return Hired Vehicle",
                "c-charge-desc_False 911 Call",
                "c-charge-desc_False Bomb Report",
                "c-charge-desc_False Imprisonment",
                "c-charge-desc_False Info LEO During Invest",
                "c-charge-desc_False Motor Veh Insurance Card",
                "c-charge-desc_False Name By Person Arrest",
                "c-charge-desc_False Ownership Info/Pawn Item",
                "c-charge-desc_Falsely Impersonating Officer",
                "c-charge-desc_Fel Drive License Perm Revoke",
                "c-charge-desc_Felon in Pos of Firearm or Amm",
                "c-charge-desc_Felony Batt(Great Bodily Harm)",
                "c-charge-desc_Felony Battery",
                "c-charge-desc_Felony Battery (Dom Strang)",
                "c-charge-desc_Felony Battery w/Prior Convict",
                "c-charge-desc_Felony Committing Prostitution",
                "c-charge-desc_Felony DUI (level 3)",
                "c-charge-desc_Felony DUI - Enhanced",
                "c-charge-desc_Felony Driving While Lic Suspd",
                "c-charge-desc_Felony Petit Theft",
                "c-charge-desc_Felony/Driving Under Influence",
                "c-charge-desc_Fighting/Baiting Animals",
                "c-charge-desc_Fleeing Or Attmp Eluding A Leo",
                "c-charge-desc_Fleeing or Eluding a LEO",
                "c-charge-desc_Forging Bank Bills/Promis Note",
                "c-charge-desc_Fraudulent Use of Credit Card",
                "c-charge-desc_Grand Theft (Motor Vehicle)",
                "c-charge-desc_Grand Theft Dwell Property",
                "c-charge-desc_Grand Theft Firearm",
                "c-charge-desc_Grand Theft in the 1st Degree",
                "c-charge-desc_Grand Theft in the 3rd Degree",
                "c-charge-desc_Grand Theft of a Fire Extinquisher",
                "c-charge-desc_Grand Theft of the 2nd Degree",
                "c-charge-desc_Grand Theft on 65 Yr or Older",
                "c-charge-desc_Harass Witness/Victm/Informnt",
                "c-charge-desc_Harm Public Servant Or Family",
                "c-charge-desc_Hiring with Intent to Defraud",
                "c-charge-desc_Imperson Public Officer or Emplyee",
                "c-charge-desc_Interfere W/Traf Cont Dev RR",
                "c-charge-desc_Interference with Custody",
                "c-charge-desc_Intoxicated/Safety Of Another",
                "c-charge-desc_Introduce Contraband Into Jail",
                "c-charge-desc_Issuing a Worthless Draft",
                "c-charge-desc_Kidnapping / Domestic Violence",
                "c-charge-desc_Lease For Purpose Trafficking",
                "c-charge-desc_Leave Acc/Attend Veh/More $50",
                "c-charge-desc_Leave Accd/Attend Veh/Less $50",
                "c-charge-desc_Leaving Acc/Unattended Veh",
                "c-charge-desc_Leaving the Scene of Accident",
                "c-charge-desc_Lewd Act Presence Child 16-",
                "c-charge-desc_Lewd or Lascivious Molestation",
                "c-charge-desc_Lewd/Lasc Battery Pers 12+/<16",
                "c-charge-desc_Lewd/Lasc Exhib Presence <16yr",
                "c-charge-desc_Lewd/Lasciv Molest Elder Persn",
                "c-charge-desc_Lewdness Violation",
                "c-charge-desc_License Suspended Revoked",
                "c-charge-desc_Littering",
                "c-charge-desc_Live on Earnings of Prostitute",
                "c-charge-desc_Lve/Scen/Acc/Veh/Prop/Damage",
                "c-charge-desc_Manage Busn W/O City Occup Lic",
                "c-charge-desc_Manslaughter W/Weapon/Firearm",
                "c-charge-desc_Manufacture Cannabis",
                "c-charge-desc_Misuse Of 911 Or E911 System",
                "c-charge-desc_Money Launder 100K or More Dols",
                "c-charge-desc_Murder In 2nd Degree W/firearm",
                "c-charge-desc_Murder in the First Degree",
                "c-charge-desc_Neglect Child / Bodily Harm",
                "c-charge-desc_Neglect Child / No Bodily Harm",
                "c-charge-desc_Neglect/Abuse Elderly Person",
                "c-charge-desc_Obstruct Fire Equipment",
                "c-charge-desc_Obstruct Officer W/Violence",
                "c-charge-desc_Obtain Control Substance By Fraud",
                "c-charge-desc_Offer Agree Secure For Lewd Act",
                "c-charge-desc_Offer Agree Secure/Lewd Act",
                "c-charge-desc_Offn Against Intellectual Prop",
                "c-charge-desc_Open Carrying Of Weapon",
                "c-charge-desc_Oper Motorcycle W/O Valid DL",
                "c-charge-desc_Operating W/O Valid License",
                "c-charge-desc_Opert With Susp DL 2nd Offens",
                "c-charge-desc_PL/Unlaw Use Credit Card",
                "c-charge-desc_Petit Theft",
                "c-charge-desc_Petit Theft $100- $300",
                "c-charge-desc_Pos Cannabis For Consideration",
                "c-charge-desc_Pos Cannabis W/Intent Sel/Del",
                "c-charge-desc_Pos Methylenedioxymethcath W/I/D/S",
                "c-charge-desc_Poss 3,4 MDMA (Ecstasy)",
                "c-charge-desc_Poss Alprazolam W/int Sell/Del",
                "c-charge-desc_Poss Anti-Shoplifting Device",
                "c-charge-desc_Poss Cntrft Contr Sub w/Intent",
                "c-charge-desc_Poss Cocaine/Intent To Del/Sel",
                "c-charge-desc_Poss Contr Subst W/o Prescript",
                "c-charge-desc_Poss Counterfeit Payment Inst",
                "c-charge-desc_Poss Drugs W/O A Prescription",
                "c-charge-desc_Poss F/Arm Delinq",
                "c-charge-desc_Poss Firearm W/Altered ID#",
                "c-charge-desc_Poss Meth/Diox/Meth/Amp (MDMA)",
                "c-charge-desc_Poss Of 1,4-Butanediol",
                "c-charge-desc_Poss Of Controlled Substance",
                "c-charge-desc_Poss Of RX Without RX",
                "c-charge-desc_Poss Oxycodone W/Int/Sell/Del",
                "c-charge-desc_Poss Pyrrolidinobutiophenone",
                "c-charge-desc_Poss Pyrrolidinovalerophenone",
                "c-charge-desc_Poss Pyrrolidinovalerophenone W/I/D/S",
                "c-charge-desc_Poss Similitude of Drivers Lic",
                "c-charge-desc_Poss Tetrahydrocannabinols",
                "c-charge-desc_Poss Unlaw Issue Driver Licenc",
                "c-charge-desc_Poss Unlaw Issue Id",
                "c-charge-desc_Poss Wep Conv Felon",
                "c-charge-desc_Poss of Cocaine W/I/D/S 1000FT Park",
                "c-charge-desc_Poss of Firearm by Convic Felo",
                "c-charge-desc_Poss of Methylethcathinone",
                "c-charge-desc_Poss/Sell/Del Cocaine 1000FT Sch",
                "c-charge-desc_Poss/Sell/Del/Man Amobarbital",
                "c-charge-desc_Poss/pur/sell/deliver Cocaine",
                "c-charge-desc_Poss3,4 Methylenedioxymethcath",
                "c-charge-desc_Posses/Disply Susp/Revk/Frd DL",
                "c-charge-desc_Possess Cannabis 1000FTSch",
                "c-charge-desc_Possess Cannabis/20 Grams Or Less",
                "c-charge-desc_Possess Controlled Substance",
                "c-charge-desc_Possess Countrfeit Credit Card",
                "c-charge-desc_Possess Drug Paraphernalia",
                "c-charge-desc_Possess Mot Veh W/Alt Vin #",
                "c-charge-desc_Possess Tobacco Product Under 18",
                "c-charge-desc_Possess Weapon On School Prop",
                "c-charge-desc_Possess w/I/Utter Forged Bills",
                "c-charge-desc_Possession Burglary Tools",
                "c-charge-desc_Possession Child Pornography",
                "c-charge-desc_Possession Firearm School Prop",
                "c-charge-desc_Possession Of 3,4Methylenediox",
                "c-charge-desc_Possession Of Alprazolam",
                "c-charge-desc_Possession Of Amphetamine",
                "c-charge-desc_Possession Of Anabolic Steroid",
                "c-charge-desc_Possession Of Buprenorphine",
                "c-charge-desc_Possession Of Carisoprodol",
                "c-charge-desc_Possession Of Clonazepam",
                "c-charge-desc_Possession Of Cocaine",
                "c-charge-desc_Possession Of Diazepam",
                "c-charge-desc_Possession Of Fentanyl",
                "c-charge-desc_Possession Of Heroin",
                "c-charge-desc_Possession Of Methamphetamine",
                "c-charge-desc_Possession Of Paraphernalia",
                "c-charge-desc_Possession Of Phentermine",
                "c-charge-desc_Possession of Alcohol Under 21",
                "c-charge-desc_Possession of Benzylpiperazine",
                "c-charge-desc_Possession of Butylone",
                "c-charge-desc_Possession of Cannabis",
                "c-charge-desc_Possession of Cocaine",
                "c-charge-desc_Possession of Codeine",
                "c-charge-desc_Possession of Ethylone",
                "c-charge-desc_Possession of Hydrocodone",
                "c-charge-desc_Possession of Hydromorphone",
                "c-charge-desc_Possession of LSD",
                "c-charge-desc_Possession of Methadone",
                "c-charge-desc_Possession of Morphine",
                "c-charge-desc_Possession of Oxycodone",
                "c-charge-desc_Possession of XLR11",
                "c-charge-desc_Principal In The First Degree",
                "c-charge-desc_Prostitution",
                "c-charge-desc_Prostitution/Lewd Act Assignation",
                "c-charge-desc_Prostitution/Lewdness/Assign",
                "c-charge-desc_Prowling/Loitering",
                "c-charge-desc_Purchase Cannabis",
                "c-charge-desc_Purchase/P/W/Int Cannabis",
                "c-charge-desc_Reckless Driving",
                "c-charge-desc_Refuse Submit Blood/Breath Test",
                "c-charge-desc_Refuse to Supply DNA Sample",
                "c-charge-desc_Resist Officer w/Violence",
                "c-charge-desc_Resist/Obstruct W/O Violence",
                "c-charge-desc_Retail Theft $300 1st Offense",
                "c-charge-desc_Retail Theft $300 2nd Offense",
                "c-charge-desc_Ride Tri-Rail Without Paying",
                "c-charge-desc_Robbery / No Weapon",
                "c-charge-desc_Robbery / Weapon",
                "c-charge-desc_Robbery Sudd Snatch No Weapon",
                "c-charge-desc_Robbery W/Deadly Weapon",
                "c-charge-desc_Robbery W/Firearm",
                "c-charge-desc_Sale/Del Cannabis At/Near Scho",
                "c-charge-desc_Sale/Del Counterfeit Cont Subs",
                "c-charge-desc_Sel/Pur/Mfr/Del Control Substa",
                "c-charge-desc_Sell or Offer for Sale Counterfeit Goods",
                "c-charge-desc_Sell/Man/Del Pos/w/int Heroin",
                "c-charge-desc_Sex Batt Faml/Cust Vict 12-17Y",
                "c-charge-desc_Sex Battery Deft 18+/Vict 11-",
                "c-charge-desc_Sex Offender Fail Comply W/Law",
                "c-charge-desc_Sexual Battery / Vict 12 Yrs +",
                "c-charge-desc_Sexual Performance by a Child",
                "c-charge-desc_Shoot In Occupied Dwell",
                "c-charge-desc_Shoot Into Vehicle",
                "c-charge-desc_Simulation of Legal Process",
                "c-charge-desc_Solic to Commit Battery",
                "c-charge-desc_Solicit Deliver Cocaine",
                "c-charge-desc_Solicit Purchase Cocaine",
                "c-charge-desc_Solicit To Deliver Cocaine",
                "c-charge-desc_Solicitation On Felony 3 Deg",
                "c-charge-desc_Soliciting For Prostitution",
                "c-charge-desc_Sound Articles Over 100",
                "c-charge-desc_Stalking",
                "c-charge-desc_Stalking (Aggravated)",
                "c-charge-desc_Strong Armed  Robbery",
                "c-charge-desc_Structuring Transactions",
                "c-charge-desc_Susp Drivers Lic 1st Offense",
                "c-charge-desc_Tamper With Victim",
                "c-charge-desc_Tamper With Witness",
                "c-charge-desc_Tamper With Witness/Victim/CI",
                "c-charge-desc_Tampering With Physical Evidence",
                "c-charge-desc_Tampering with a Victim",
                "c-charge-desc_Theft/To Deprive",
                "c-charge-desc_Threat Public Servant",
                "c-charge-desc_Throw Deadly Missile Into Veh",
                "c-charge-desc_Throw In Occupied Dwell",
                "c-charge-desc_Throw Missile Into Pub/Priv Dw",
                "c-charge-desc_Traff In Cocaine <400g>150 Kil",
                "c-charge-desc_Traffic Counterfeit Cred Cards",
                "c-charge-desc_Traffick Amphetamine 28g><200g",
                "c-charge-desc_Traffick Oxycodone     4g><14g",
                "c-charge-desc_Trans/Harm/Material to a Minor",
                "c-charge-desc_Trespass On School Grounds",
                "c-charge-desc_Trespass Other Struct/Conve",
                "c-charge-desc_Trespass Private Property",
                "c-charge-desc_Trespass Property w/Dang Weap",
                "c-charge-desc_Trespass Struct/Conveyance",
                "c-charge-desc_Trespass Structure w/Dang Weap",
                "c-charge-desc_Trespass Structure/Conveyance",
                "c-charge-desc_Trespassing/Construction Site",
                "c-charge-desc_Tresspass Struct/Conveyance",
                "c-charge-desc_Tresspass in Structure or Conveyance",
                "c-charge-desc_Unauth C/P/S Sounds>1000/Audio",
                "c-charge-desc_Unauth Poss ID Card or DL",
                "c-charge-desc_Unauthorized Interf w/Railroad",
                "c-charge-desc_Unl/Disturb Education/Instui",
                "c-charge-desc_Unlaw Lic Use/Disply Of Others",
                "c-charge-desc_Unlaw LicTag/Sticker Attach",
                "c-charge-desc_Unlaw Use False Name/Identity",
                "c-charge-desc_Unlawful Conveyance of Fuel",
                "c-charge-desc_Unlicensed Telemarketing",
                "c-charge-desc_Use Computer for Child Exploit",
                "c-charge-desc_Use Of 2 Way Device To Fac Fel",
                "c-charge-desc_Use Scanning Device to Defraud",
                "c-charge-desc_Use of Anti-Shoplifting Device",
                "c-charge-desc_Uttering Forged Bills",
                "c-charge-desc_Uttering Forged Credit Card",
                "c-charge-desc_Uttering Worthless Check +$150",
                "c-charge-desc_Uttering a Forged Instrument",
                "c-charge-desc_Video Voyeur-<24Y on Child >16",
                "c-charge-desc_Viol Injunct Domestic Violence",
                "c-charge-desc_Viol Injunction Protect Dom Vi",
                "c-charge-desc_Viol Pretrial Release Dom Viol",
                "c-charge-desc_Viol Prot Injunc Repeat Viol",
                "c-charge-desc_Violation License Restrictions",
                "c-charge-desc_Violation Of Boater Safety Id",
                "c-charge-desc_Violation of Injunction Order/Stalking/Cyberstalking",
                "c-charge-desc_Voyeurism",
                "c-charge-desc_arrest case no charge",
            ],
        }
        discrete_features = flatten_dict(disc_feature_groups)

        continuous_features = [
            "age-num",
            "juv-fel-count",
            "juv-misd-count",
            "juv-other-count",
            "priors-count",
        ]

        sens_attr_spec: Union[str, LabelSpec]
        if self.split is CompasSplits.SEX:
            sens_attr_spec = "sex"
            s_prefix = ["sex"]
            class_label_spec = "two-year-recid"
            class_label_prefix = ["two-year-recid"]
        elif self.split is CompasSplits.RACE:
            sens_attr_spec = "race"
            s_prefix = ["race"]
            class_label_spec = "two-year-recid"
            class_label_prefix = ["two-year-recid"]
        elif self.split is CompasSplits.RACE_SEX:
            sens_attr_spec = simple_spec({"sex": ["sex"], "race": ["race"]})
            s_prefix = ["race", "sex"]
            class_label_spec = "two-year-recid"
            class_label_prefix = ["two-year-recid"]
        elif self.split is CompasSplits.CUSTOM:
            sens_attr_spec = ""
            s_prefix = []
            class_label_spec = ""
            class_label_prefix = []
        else:
            raise NotImplementedError

        super().__init__(
            name=f"Compas {self.split.value}",
            num_samples=6167,
            filename_or_path="compas-recidivism.csv",
            features=discrete_features + continuous_features,
            cont_features=continuous_features,
            s_prefix=s_prefix,
            sens_attr_spec=sens_attr_spec,
            class_label_prefix=class_label_prefix,
            class_label_spec=class_label_spec,
            discrete_only=self.discrete_only,
            discrete_feature_groups=disc_feature_groups,
        )



"""Implementation of Agarwal model."""
from dataclasses import dataclass
from typing import List, Optional, Set
from typing_extensions import TypedDict

from ranzen import implements

from ethicml.algorithms.inprocess.in_subprocess import InAlgorithmSubprocess
from ethicml.algorithms.inprocess.shared import settings_for_svm_lr
from ethicml.utility import ClassifierType, FairnessType, HyperParamType

__all__ = ["Agarwal"]

from ethicml.utility import KernelType

VALID_MODELS: Set[ClassifierType] = {ClassifierType.lr, ClassifierType.svm}


class AgarwalArgs(TypedDict):
    """Args for the Agarwal implementation."""

    classifier: str
    fairness: str
    eps: float
    iters: int
    C: float
    kernel: str


@dataclass
class Agarwal(InAlgorithmSubprocess):
    """Agarwal class.

    A wrapper around the Exponentiated Gradient method documented
    `on this website <https://fairlearn.org/v0.7.0/api_reference/fairlearn.reductions.html#fairlearn.reductions.ExponentiatedGradient>`_.

    :param fairness: Type of fairness to enforce.
    :param classifier: Type of classifier to use.
    :param eps: Epsilon fo.
    :param iters: Number of iterations for the DP algorithm.
    :param C: C parameter for the SVM algorithm.
    :param kernel: Kernel type for the SVM algorithm.
    """

    fairness: FairnessType = FairnessType.dp
    classifier: ClassifierType = ClassifierType.lr
    eps: float = 0.1
    iters: int = 50
    C: Optional[float] = None
    kernel: Optional[KernelType] = None

    def __post_init__(self) -> None:
        assert self.fairness in (FairnessType.dp, FairnessType.eq_odds)

    @implements(InAlgorithmSubprocess)
    def _get_flags(self) -> AgarwalArgs:
        chosen_c, chosen_kernel = settings_for_svm_lr(self.classifier, self.C, self.kernel)
        # TODO: replace this with dataclasses.asdict()
        return {
            "classifier": str(self.classifier),
            "fairness": str(self.fairness),
            "eps": self.eps,
            "iters": self.iters,
            "C": chosen_c,
            "kernel": str(chosen_kernel) if chosen_kernel is not None else "",
        }

    @implements(InAlgorithmSubprocess)
    def get_hyperparameters(self) -> HyperParamType:
        chosen_c, chosen_kernel = settings_for_svm_lr(self.classifier, self.C, self.kernel)
        _hyperparameters: HyperParamType = {
            "C": chosen_c,
            "iters": self.iters,
            "eps": self.eps,
            "fairness": self.fairness,
        }
        if self.classifier is ClassifierType.svm:
            assert chosen_kernel is not None
            _hyperparameters["kernel"] = chosen_kernel
        return _hyperparameters

    @implements(InAlgorithmSubprocess)
    def get_name(self) -> str:
        return f"Agarwal, {self.classifier}, {self.fairness}, {self.eps}"

    @implements(InAlgorithmSubprocess)
    def _get_path_to_script(self) -> List[str]:
        return ["-m", "ethicml.implementations.agarwal"]



"""Fair Dummies utility functions."""
from typing import List, Optional, Tuple

import numpy as np
from sklearn.neighbors import KernelDensity


def density_estimation(
    y: np.ndarray, *, a: np.ndarray, y_test: Optional[np.ndarray] = None
) -> Tuple[List[float], List[float]]:
    """Estimate the distribusion of P{A|Y}."""
    if y_test is None:
        y_test = np.array([])
    assert y_test is not None
    bandwidth = np.sqrt(max(np.median(np.abs(y)), 0.01))

    kde_0 = KernelDensity(kernel='linear', bandwidth=bandwidth).fit(y[a == 0][:, np.newaxis])
    kde_1 = KernelDensity(kernel='linear', bandwidth=bandwidth).fit(y[a == 1][:, np.newaxis])

    log_dens_0 = np.exp(np.squeeze(kde_0.score_samples(y[:, np.newaxis])))
    log_dens_1 = np.exp(np.squeeze(kde_1.score_samples(y[:, np.newaxis])))
    p_0 = np.sum(a == 0) / a.shape[0]
    p_1 = 1 - p_0

    # p(A=1|y) = p(y|A=1)p(A=1) / (p(y|A=1)p(A=1) + p(y|A=0)p(A=0))
    p_success = (log_dens_1 * p_1) / (log_dens_1 * p_1 + log_dens_0 * p_0 + 1e-10)

    p_success_test = []
    if len(y_test) > 0:
        log_dens_0_test = np.exp(np.squeeze(kde_0.score_samples(y_test[:, np.newaxis])))
        log_dens_1_test = np.exp(np.squeeze(kde_1.score_samples(y_test[:, np.newaxis])))
        p_success_test = (log_dens_1_test * p_1) / (
            log_dens_1_test * p_1 + log_dens_0_test * p_0 + 1e-10
        )

    return p_success, p_success_test



"""Implementation for Louizos et al Variational Fair Autoencoder."""
# pylint: disable=arguments-differ

from typing import Any, List, NamedTuple, Optional, Tuple

import torch
from torch import Tensor, nn

from ethicml.data.dataset import Dataset

from .decoder import Decoder
from .encoder import Encoder


class LvInfo(NamedTuple):
    """Hold Latent Variable Information."""

    embed: Tensor
    mean: Tensor
    logvar: Tensor


class VFAENetwork(nn.Module):
    """Implements a generative model with two layers of stochastic variables.

    Where both are conditional, i.e.:

    p(x, z1, z2, y | s) = p(z2) p(y) p(z1 | z2, y) p(x | z1, s)

    with q(z1 | x, s) q(z2 | z1, y) q(y | z1) being the variational posteriors.
    """

    def __init__(
        self,
        dataset: Dataset,
        supervised: bool,
        input_size: int,
        latent_dims: int,
        z1_enc_size: List[int],
        z2_enc_size: List[int],
        z1_dec_size: List[int],
    ):
        super().__init__()
        torch.manual_seed(888)

        self.supervised = supervised

        self.z1_encoder = Encoder(z1_enc_size, input_size + 1, latent_dims)
        if self.supervised:
            self.z2_encoder = Encoder(z2_enc_size, latent_dims + 1, latent_dims)
            self.z1_decoder = Encoder(z1_dec_size, latent_dims + 1, latent_dims)
        self.x_dec = Decoder(dataset)
        self.ypred = nn.Linear(latent_dims, 1)

    def encode_z1(self, x: Tensor, s: Tensor) -> Tuple[Tensor, Tensor]:
        """Encode Z1.

        :param x:
        :param s:
        """
        return self.z1_encoder(torch.cat((x, s.view(-1, 1)), 1))

    def encode_z2(self, z1: Tensor, y: Tensor) -> Tuple[Tensor, Tensor]:
        """Encode Z2.

        :param z1:
        :param y:
        """
        return self.z2_encoder(torch.cat((z1, y.view(-1, 1)), 1))

    def decode_z1(self, z2: Tensor, y: Tensor) -> Tuple[Tensor, Tensor]:
        """Decode Z1.

        :param z2:
        :param y:
        """
        return self.z1_decoder(torch.cat((z2, y.view(-1, 1)), 1))

    @staticmethod
    def reparameterize(mean: Tensor, logvar: Tensor) -> Tensor:
        """Reparametrization trick - Leaving as a method to try and control reproducability.

        :param mean:
        :param logvar:
        """
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return eps.mul(std).add_(mean)

    def forward(
        self, x: Tensor, s: Tensor, y: Tensor
    ) -> Tuple[LvInfo, Optional[LvInfo], Optional[LvInfo], Tensor, Optional[Tensor],]:
        """Forward pass for network.

        :param x:
        :param s:
        :param y:
        """
        z1_mu, z1_logvar = self.encode_z1(x, s)
        z1 = self.reparameterize(z1_mu, z1_logvar)

        z2_triplet: Optional[Tuple[Any, Any, Any]]
        z1_d_triplet: Optional[Tuple[Any, Any, Any]]
        y_pred: Optional[torch.Tensor]
        if self.supervised:
            z2_mu, z2_logvar = self.encode_z2(z1, y)
            z2 = self.reparameterize(z2_mu, z2_logvar)

            z1_dec_mu, z1_dec_logvar = self.decode_z1(z2, y)
            z1_dec = self.reparameterize(z1_dec_mu, z1_dec_logvar)

            x_dec = self.x_dec(z1_dec, s)

            y_pred = torch.sigmoid(self.ypred(z1))

            z2_triplet = LvInfo(embed=z2, mean=z2_mu, logvar=z2_logvar)
            z1_d_triplet = LvInfo(embed=z1_dec, mean=z1_dec_mu, logvar=z1_dec_logvar)

        else:
            x_dec = self.x_dec(z1, s)
            z2_triplet = None
            z1_d_triplet = None
            y_pred = None

        z1_triplet = LvInfo(embed=z1, mean=z1_mu, logvar=z1_logvar)

        return z1_triplet, z2_triplet, z1_d_triplet, x_dec, y_pred



"""Script to generate the toy dataset."""
import random
from pathlib import Path

import numpy as np
import pandas as pd


def main(seed: int, samples: int) -> None:
    """Make a toy dataset that can be used to check an algorithm runs."""
    np.random.seed(seed)
    random.seed(seed)

    feat1_g1: np.ndarray = np.random.randn(samples // 2) - 0.1
    feat1_g2: np.ndarray = np.random.randn(samples // 2) + 0.5

    feat2_g1: np.ndarray = np.random.randn(samples // 2) * 0.9
    feat2_g2: np.ndarray = np.random.randn(samples // 2) * 1.1

    disc_1 = np.random.choice(['a', 'b', 'c', 'd', 'e'], [samples], replace=True)
    disc_2 = np.random.choice(['x', 'y', 'z'], [samples], replace=True)

    sensitive_attr_g1 = np.full(feat1_g1.shape, 0)
    sensitive_attr_g2 = np.full(feat1_g2.shape, 1)
    feat1 = np.concatenate((feat1_g1, feat1_g2))
    feat2 = np.concatenate((feat2_g1, feat2_g2))
    sensitive_attr = np.concatenate((sensitive_attr_g1, sensitive_attr_g2))

    # There has to be a better way of dong this.
    decision = np.array([1 if v > 0 else 0 for v in feat1 + feat2 > 0])
    df = pd.DataFrame(
        data={
            "decision": decision,
            "sensitive-attr": sensitive_attr,
            "a1": feat1,
            "a2": feat2,
            "disc_1": disc_1,
            "disc_2": disc_2,
        }
    )
    df = pd.get_dummies(df)

    # Shuffle the data,
    df = df.sample(frac=1.0, random_state=seed).reset_index(drop=True)

    # Save the CSV
    df.to_csv(str(Path(__file__).parent / "toy.csv"), index=False)


if __name__ == "__main__":
    SEED = 888
    SAMPLES = 400
    main(SEED, SAMPLES)



"""Collection of functions that enable parallelism."""
from typing import List, Optional, Sequence, Tuple, TypeVar, Union, cast, overload
from typing_extensions import Protocol, TypeAlias

import numpy as np
import pandas as pd
from joblib import Parallel, delayed

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm
from ethicml.algorithms.preprocess.pre_algorithm import PreAlgorithm
from ethicml.utility.data_structures import DataTuple, Prediction, SubgroupTuple, TrainValPair

__all__ = ["arrange_in_parallel", "run_in_parallel"]


InSeq: TypeAlias = Sequence[InAlgorithm]
PreSeq: TypeAlias = Sequence[PreAlgorithm]
InResult: TypeAlias = List[List[Prediction]]
PreResult: TypeAlias = List[List[Tuple[DataTuple, DataTuple]]]
DataSeq: TypeAlias = Sequence[TrainValPair]


@overload
def run_in_parallel(
    algos: InSeq, *, data: DataSeq, seeds: List[int], num_jobs: Optional[int] = None
) -> InResult:
    ...


@overload
def run_in_parallel(
    algos: PreSeq, *, data: DataSeq, seeds: List[int], num_jobs: Optional[int] = None
) -> PreResult:
    ...


def run_in_parallel(
    algos: Union[InSeq, PreSeq], *, data: DataSeq, seeds: List[int], num_jobs: Optional[int] = None
) -> Union[InResult, PreResult]:
    """Run the given algorithms (embarrassingly) parallel.

    :param algos: list of algorithms
    :param data: list of pairs of data tuples (train and test)
    :param seeds: list of seeds to use when running the model
    :param num_jobs: how many jobs can run in parallel at most. if None, use number of CPUs (Default: None)
    :returns: list of the results
    """
    if not algos or not data:
        return cast(List[List[Prediction]], [[]])
    # The following isinstance check is not at all reliable because `InAlgorithm` is a Protocol,
    # but that's completely fine because this check is only here for mypy anyway.
    if isinstance(algos[0], InAlgorithm):
        in_algos = cast(Sequence[InAlgorithm], algos)
        return arrange_in_parallel(algos=in_algos, data=data, seeds=seeds, num_jobs=num_jobs)
    else:
        pre_algos = cast(Sequence[PreAlgorithm], algos)
        # the following line is needed to help mypy along
        generic_algos: Sequence[Algorithm[Tuple[DataTuple, DataTuple]]] = pre_algos
        return arrange_in_parallel(algos=generic_algos, data=data, seeds=seeds, num_jobs=num_jobs)


_RT = TypeVar("_RT", Prediction, Tuple[DataTuple, DataTuple], covariant=True)  # the return type


class Algorithm(Protocol[_RT]):
    """Protocol for making `arrange_in_parallel` generic."""

    def run(self, train: DataTuple, test: DataTuple, seed: int) -> _RT:
        ...


def arrange_in_parallel(
    algos: Sequence[Algorithm[_RT]], data: DataSeq, seeds: List[int], num_jobs: Optional[int] = None
) -> List[List[_RT]]:
    """Arrange the given algorithms to run (embarrassingly) parallel.

    :param algos: list of tuples consisting of a `run_async` function of an algorithm and a name
    :param data: list of pairs of data tuples (train and test)
    :param num_jobs: number of parallel jobs. `None` means as many as available CPUs.
        (Default: None)
    :returns: list of the results
    """
    runner = Parallel(n_jobs=num_jobs, verbose=10, backend="loky")
    assert len(algos) >= 1
    assert len(data) >= 1
    assert len(seeds) == len(data)
    assert isinstance(data[0][0], DataTuple)
    assert isinstance(data[0][1], (DataTuple, SubgroupTuple))
    # ================================== create queue of tasks ====================================
    # for each algorithm, first loop over all available datasets and then go on to the next algo
    results = runner(
        _run(algo, data_item, seed) for algo in algos for (data_item, seed) in zip(data, seeds)
    )
    # return [[result_dict[(i, j)] for j in range(len(data))] for i in range(len(algos))]
    # we have to reconstruct the nested list from the flattened list
    return [[results[i * len(data) + j] for j in range(len(data))] for i in range(len(algos))]


@delayed
def _run(algo: Algorithm[_RT], train_test_pair: TrainValPair, seed: int) -> _RT:
    train, test = train_test_pair
    # do the work
    try:
        result: _RT = algo.run(train, test, seed)
    except RuntimeError:
        result = Prediction(hard=pd.Series([np.NaN] * len(test)))
    if isinstance(result, Prediction):
        result.info["model_seed"] = seed
    return result



"""Implementation of Fairness without Demographics."""
from __future__ import annotations

import json
import sys
from pathlib import Path
from typing import TYPE_CHECKING, List, Union

import numpy as np
from joblib import dump, load

try:
    import torch
    from torch import optim
    from torch.optim.optimizer import Optimizer
    from torch.utils.data import DataLoader

except ImportError as e:
    raise RuntimeError(
        "In order to use PyTorch, please install it following the instructions at https://pytorch.org/ . "
    ) from e


from joblib import dump, load

from ethicml.implementations.beutel import set_seed
from ethicml.implementations.dro_modules.dro_classifier import DROClassifier
from ethicml.implementations.pytorch_common import CustomDataset, TestDataset
from ethicml.implementations.utils import load_data_from_flags
from ethicml.utility import DataTuple, SoftPrediction, SubgroupTuple, TestTuple

if TYPE_CHECKING:
    from ethicml.algorithms.inprocess.fairness_wo_demographics import DroArgs
    from ethicml.algorithms.inprocess.in_subprocess import InAlgoArgs


def train_model(
    epoch: int, model: DROClassifier, train_loader: DataLoader, optimizer: Optimizer
) -> None:
    """Train a model.

    :param epoch:
    :param model:
    :param train_loader:
    :param optimizer:
    """
    model.train()
    train_loss = 0.0
    for batch_idx, (data_x, _, data_y) in enumerate(train_loader):
        data_x = data_x.to("cpu")
        data_y = data_y.to("cpu")
        optimizer.zero_grad()
        y_prob = model.forward(data_x)
        loss = model.loss(y_prob.squeeze(-1), data_y)
        loss.backward()
        optimizer.step()
        train_loss += loss.item()
        if batch_idx % 100 == 0:
            print(
                f"train Epoch: {epoch} [{batch_idx * len(data_x)}/{len(train_loader.dataset)}"
                f"\t({100. * batch_idx / len(train_loader):.0f}%)]"
                f"\tLoss: {loss.item() / len(data_x):.6f}"
            )

    print(f"====> Epoch: {epoch} Average loss: {train_loss / len(train_loader.dataset):.4f}")


def fit(train: DataTuple, args: DroArgs, seed: int) -> DROClassifier:
    """Train a network and return predictions.

    :param train:
    :param args:
    """
    # Set up the data
    set_seed(seed)
    train_data = CustomDataset(train)
    train_loader = DataLoader(train_data, batch_size=args["batch_size"])

    # Build Network
    model = DROClassifier(
        in_size=train_data.xdim,
        out_size=train_data.ydim,
        network_size=args["network_size"],
        eta=args["eta"],
    ).to("cpu")
    optimizer = optim.Adam(model.parameters(), lr=1e-3)

    # Run Network
    for epoch in range(int(args["epochs"])):
        train_model(epoch, model, train_loader, optimizer)
    return model


def predict(model: DROClassifier, test: TestTuple, args: DroArgs) -> SoftPrediction:
    """Train a network and return predictions.

    :param model:
    :param test:
    :param args:
    """
    # Set up the data
    test_data = TestDataset(test)
    test_loader = DataLoader(test_data, batch_size=args["batch_size"])

    # Transform output
    post_test: List[torch.Tensor] = []
    model.eval()
    with torch.no_grad():
        for _x, _ in test_loader:
            post_test += model.forward(_x)

    return SoftPrediction(soft=np.array([j.softmax(dim=-1).numpy() for j in post_test]))


def train_and_predict(
    train: DataTuple, test: TestTuple, args: DroArgs, seed: int
) -> SoftPrediction:
    """Train a network and return predictions.

    :param train:
    :param test:
    :param args:
    """
    # Set up the data
    set_seed(seed)
    train_data = CustomDataset(train)
    train_loader = DataLoader(train_data, batch_size=args["batch_size"])

    test_data = TestDataset(test)
    test_loader = DataLoader(test_data, batch_size=args["batch_size"])

    # Build Network
    model = DROClassifier(
        in_size=train_data.xdim,
        out_size=train_data.ydim,
        network_size=args["network_size"],
        eta=args["eta"],
    ).to("cpu")
    optimizer = optim.Adam(model.parameters(), lr=1e-3)

    # Run Network
    for epoch in range(int(args["epochs"])):
        train_model(epoch, model, train_loader, optimizer)

    # Transform output
    post_test: List[torch.Tensor] = []
    model.eval()
    with torch.no_grad():
        for _x, _ in test_loader:
            post_test += model.forward(_x)

    return SoftPrediction(soft=np.array([j.softmax(dim=-1).numpy() for j in post_test]))


def main() -> None:
    """Run the FWD model as a standalone program on tabular data."""
    in_algo_args: InAlgoArgs = json.loads(sys.argv[1])
    flags: DroArgs = json.loads(sys.argv[2])
    data: Union[DataTuple, TestTuple]
    if in_algo_args["mode"] == "run":
        train, test = load_data_from_flags(in_algo_args)
        train_and_predict(train, test, flags, seed=in_algo_args["seed"]).to_npz(
            Path(in_algo_args["predictions"])
        )
    elif in_algo_args["mode"] == "fit":
        data = DataTuple.from_npz(Path(in_algo_args["train"]))
        model = fit(data, flags, seed=in_algo_args["seed"])
        dump(model, Path(in_algo_args["model"]))
    elif in_algo_args["mode"] == "predict":
        data = SubgroupTuple.from_npz(Path(in_algo_args["test"]))
        model = load(Path(in_algo_args["model"]))
        predict(model, data, flags).to_npz(Path(in_algo_args["predictions"]))


if __name__ == "__main__":
    main()



"""Useful functions used in implementations."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Tuple

from ethicml.utility import DataTuple, SubgroupTuple, TestTuple

if TYPE_CHECKING:
    from ethicml.algorithms.inprocess.in_subprocess import InAlgoRunArgs
    from ethicml.algorithms.preprocess.pre_subprocess import PreAlgoRunArgs


def load_data_from_flags(args: PreAlgoRunArgs | InAlgoRunArgs) -> Tuple[DataTuple, SubgroupTuple]:
    """Load data from the paths specified in the flags.

    :param args:
    """
    return DataTuple.from_npz(Path(args["train"])), SubgroupTuple.from_npz(Path(args["test"]))


def save_transformations(
    transforms: Tuple[DataTuple, TestTuple], pre_algo_run_args: PreAlgoRunArgs
) -> None:
    """Save the data to the file that was specified in the commandline arguments.

    :param transforms:
    :param pre_algo_run_args:
    """
    train, test = transforms
    assert isinstance(train, DataTuple)
    assert isinstance(test, (DataTuple, SubgroupTuple))
    train.to_npz(Path(pre_algo_run_args["new_train"]))
    test.to_npz(Path(pre_algo_run_args["new_test"]))



"""Anti-spurious."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

from .metric import Metric, MetricStaticName

__all__ = ["AS"]


@dataclass
class AS(MetricStaticName):
    r"""Anti-spurious metric.

    Computes :math:`P(\hat{y}=y|y\neq s)`.
    """

    _name: ClassVar[str] = "anti_spurious"

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        preds = prediction.hard.to_numpy()
        sens = actual.s.to_numpy()
        labels = actual.y.to_numpy()
        s_uneq_y = sens != labels
        return (preds[s_uneq_y] == labels[s_uneq_y]).mean()



"""Make the law admissions dataset."""
import pandas as pd


def main() -> None:
    """Make the Law Dataset."""
    data = pd.read_csv("raw/law_data.csv", index_col=0).reset_index(drop=True)
    data = data[["LSAT", "UGPA", "ZFYA", "race", "sex", "first_pf"]]
    data["first_pf"] = data["first_pf"].astype(int)
    data = data.rename(columns={"race": "Race", "sex": "Sex", "first_pf": "PF"})
    data = pd.get_dummies(data, columns=["Race", "Sex", "PF"])
    data = data.sample(frac=1.0, random_state=888).reset_index(drop=True)
    compression_opts = dict(method='zip', archive_name='law.csv')
    print(data.head())
    data.to_csv("./law.csv.zip", index=False, compression=compression_opts)


if __name__ == '__main__':
    main()



"""Implementation for Louizos et al Variational Fair Autoencoder."""

from typing import Optional, Tuple

import torch
import torch.nn.functional as F
from torch import Tensor

from ethicml.algorithms.preprocess.vfae import VfaeArgs

from ..pytorch_common import quadratic_time_mmd
from .vfae_network import LvInfo

__all__ = ["kullback_leibler", "loss_function"]

from ethicml.utility import FairnessType


def kullback_leibler(
    mu1: Tensor, logvar1: Tensor, mu2: Optional[Tensor] = None, logvar2: Optional[Tensor] = None
) -> Tensor:
    """KL Divergence.

    :param mu1:
    :param logvar1:
    :param mu2:  (Default: None)
    :param logvar2:  (Default: None)
    :returns: Tensorof divergence in each dim.
    """
    mu2 = mu2 if mu2 is not None else torch.tensor([0.0])
    logvar2 = logvar2 if logvar2 is not None else torch.tensor([0.0])
    return (
        0.5
        * (logvar2 - logvar1 - 1 + ((logvar1).exp() + ((mu1 - mu2) ** 2)) / (logvar2).exp()).sum()
    )


def loss_function(
    flags: VfaeArgs,
    z1_triplet: LvInfo,
    z2_triplet: Optional[LvInfo],
    z1_d_triplet: Optional[LvInfo],
    data_triplet: Tuple[Tensor, Tensor, Tensor],
    x_dec: Tensor,
    y_pred: Optional[Tensor],
) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
    """Loss function for VFAE.

    :param flags:
    :param z1_triplet:
    :param z2_triplet:
    :param z1_d_triplet:
    :param data_triplet:
    :param x_dec:
    :param y_pred:
    :returns: Tuple of prediction loss, reconstruction loss, KL Divergence and MMD.
    """
    z1, z1_mu, z1_logvar = z1_triplet
    if flags["supervised"]:
        assert z2_triplet is not None
        assert z1_d_triplet is not None
        _, z2_mu, z2_logvar = z2_triplet
        _, z1_dec_mu, z1_dec_logvar = z1_d_triplet
    x, s, y = data_triplet
    s = s.view(-1, 1)
    y = y.view(-1, 1)

    reconstruction_loss = F.mse_loss(x_dec, x, reduction="sum")

    fairness = FairnessType[flags["fairness"]]
    if fairness is FairnessType.dp:
        z1_s0 = torch.masked_select(z1, s.le(0.5)).view(-1, 50)
        z1_s1 = torch.masked_select(z1, s.ge(0.5)).view(-1, 50)
    elif fairness is FairnessType.eq_opp:
        z1_s0 = torch.masked_select(z1, s.le(0.5)).view(-1, 50)
        y_s0 = torch.masked_select(y, s.le(0.5)).view(-1, 1)
        z1_s0_y1 = torch.masked_select(z1_s0, y_s0.ge(0.5)).view(-1, 50)
        z1_s1 = torch.masked_select(z1, s.ge(0.5)).view(-1, 50)
        y_s1 = torch.masked_select(y, s.ge(0.5)).view(-1, 1)
        # assert False, f"{z1_s1.shape}, {y_s1.shape}"
        z1_s1_y1 = torch.masked_select(z1_s1, y_s1.ge(0.5)).view(-1, 50)

        z1_s0 = z1_s0_y1
        z1_s1 = z1_s1_y1
    else:
        raise NotImplementedError(
            f"Only {FairnessType.dp} and {FairnessType.eq_opp} implementesd so far"
        )

    mmd_loss = quadratic_time_mmd(z1_s0, z1_s1, 2.5)

    if flags["supervised"]:
        first_kl = kullback_leibler(z2_mu, z2_logvar)
        second_kl = kullback_leibler(z1_dec_mu, z1_dec_logvar, z1_mu, z1_logvar)
        kl_div = first_kl + second_kl
        assert y_pred is not None
        prediction_loss = F.binary_cross_entropy(y_pred, y, reduction="sum")
    else:
        kl_div = torch.zeros(1)
        prediction_loss = torch.zeros(1)

    return prediction_loss, reconstruction_loss, kl_div, 100 * mmd_loss



"""Fairness without Demographics."""

from dataclasses import asdict, dataclass, field
from typing import List
from typing_extensions import TypedDict

from ranzen import implements

from ethicml.algorithms.inprocess.in_subprocess import InAlgorithmSubprocess
from ethicml.utility import HyperParamType

__all__ = ["DRO"]


class DroArgs(TypedDict):
    """Args used in this module."""

    batch_size: int
    epochs: int
    eta: float
    network_size: List[int]


@dataclass
class DRO(InAlgorithmSubprocess):
    """Implementation of https://arxiv.org/abs/1806.08010 .

    :param eta: Tolerance.
    :param epochs: The number of epochs to train for.
    :param batch_size: The batch size.
    :param network_size: The size of the network.
    """

    eta: float = 0.5
    epochs: int = 10
    batch_size: int = 32
    network_size: List[int] = field(default_factory=lambda: [50])

    @implements(InAlgorithmSubprocess)
    def _get_flags(self) -> DroArgs:
        # TODO: replace this with dataclasses.asdict()
        return {
            "eta": self.eta,
            "batch_size": self.batch_size,
            "epochs": self.epochs,
            "network_size": self.network_size,
        }

    @implements(InAlgorithmSubprocess)
    def get_hyperparameters(self) -> HyperParamType:
        _hyperparameters = asdict(self)
        _hyperparameters.pop("dir")  # this is not really a hyperparameter
        return _hyperparameters

    @implements(InAlgorithmSubprocess)
    def get_name(self) -> str:
        return "Dist Robust Optim"

    @implements(InAlgorithmSubprocess)
    def _get_path_to_script(self) -> List[str]:
        return ["-m", "ethicml.implementations.dro_tabular"]



"""Class to describe features of the SQF dataset."""
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type, Union

from ..dataset import LoadableDataset
from ..util import LabelSpec, flatten_dict, simple_spec

__all__ = ["Sqf", "SqfSplits"]


class SqfSplits(Enum):
    """Splits for the SQF dataset."""

    SEX = "Sex"
    RACE = "Race"
    RACE_SEX = "Race-Sex"
    CUSTOM = "Custom"


@dataclass
class Sqf(LoadableDataset):
    """Stop, question and frisk dataset.

    This data is from the 2016, source: http://www1.nyc.gov/site/nypd/stats/reports-analysis/stopfrisk.page
    """

    split: SqfSplits = SqfSplits.SEX

    Splits: ClassVar[Type[SqfSplits]] = SqfSplits
    """Shorthand for the Enum that defines the splits associated with this class."""

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "sex": ["sex"],
            "race": ["race"],
            "weapon": ["weapon"],
            "typeofid": ["typeofid_O", "typeofid_P", "typeofid_R", "typeofid_V"],
            "explnstp": ["explnstp_N", "explnstp_Y"],
            "othpers": ["othpers_N", "othpers_Y"],
            "arstmade": ["arstmade_N", "arstmade_Y"],
            "offunif": ["offunif_N", "offunif_Y"],
            "frisked": ["frisked_N", "frisked_Y"],
            "searched": ["searched_N", "searched_Y"],
            "pf-hands": ["pf-hands_N", "pf-hands_Y"],
            "pf-wall": ["pf-wall_N", "pf-wall_Y"],
            "pf-grnd": ["pf-grnd_N", "pf-grnd_Y"],
            "pf-drweb": ["pf-drwep_N", "pf-drwep_Y"],
            "pf-ptwep": ["pf-ptwep_N", "pf-ptwep_Y"],
            "pf-baton": ["pf-baton_N", "pf-baton_Y"],
            "pf-hcuff": ["pf-hcuff_N", "pf-hcuff_Y"],
            "pf-pepsp": ["pf-pepsp_N", "pf-pepsp_Y"],
            "pf-other": ["pf-other_N", "pf-other_Y"],
            "radio": ["radio_N", "radio_Y"],
            "ac-rept": ["ac-rept_N", "ac-rept_Y"],
            "ac-inves": ["ac-inves_N", "ac-inves_Y"],
            "rf-vcrim": ["rf-vcrim_N", "rf-vcrim_Y"],
            "rf-othsw": ["rf-othsw_N", "rf-othsw_Y"],
            "ac-proxm": ["ac-proxm_N", "ac-proxm_Y"],
            "rf-attir": ["rf-attir_N", "rf-attir_Y"],
            "cs-objcs": ["cs-objcs_N", "cs-objcs_Y"],
            "cs-descr": ["cs-descr_N", "cs-descr_Y"],
            "cs-casng": ["cs-casng_N", "cs-casng_Y"],
            "cs-lkout": ["cs-lkout_N", "cs-lkout_Y"],
            "rf-vacnt": ["rf-vcact_N", "rf-vcact_Y"],
            "cs-cloth": ["cs-cloth_N", "cs-cloth_Y"],
            "cs-drgtr": ["cs-drgtr_N", "cs-drgtr_Y"],
            "ac-evasv": ["ac-evasv_N", "ac-evasv_Y"],
            "ac-assos": ["ac-assoc_N", "ac-assoc_Y"],
            "cs-furtv": ["cs-furtv_N", "cs-furtv_Y"],
            "rf-rfcmp": ["rf-rfcmp_N", "rf-rfcmp_Y"],
            "ac-cgdir": ["ac-cgdir_N", "ac-cgdir_Y"],
            "rf-verbl": ["rf-verbl_N", "rf-verbl_Y"],
            "cs-vcrim": ["cs-vcrim_N", "cs-vcrim_Y"],
            "cs-bulge": ["cs-bulge_N", "cs-bulge_Y"],
            "cs-other": ["cs-other_N", "cs-other_Y"],
            "ac-incid": ["ac-incid_N", "ac-incid_Y"],
            "ac-time": ["ac-time_N", "ac-time_Y"],
            "rf-knowl": ["rf-knowl_N", "rf-knowl_Y"],
            "ac-stsnd": ["ac-stsnd_N", "ac-stsnd_Y"],
            "ac-other": ["ac-other_N", "ac-other_Y"],
            "sb-hdobj": ["sb-hdobj_N", "sb-hdobj_Y"],
            "sb-outln": ["sb-outln_N", "sb-outln_Y"],
            "sb-admis": ["sb-admis_N", "sb-admis_Y"],
            "sb-other": ["sb-other_N", "sb-other_Y"],
            "rf-furt": ["rf-furt_N", "rf-furt_Y"],
            "rf-bulg": ["rf-bulg_N", "rf-bulg_Y"],
            "offverb": ["offverb_ ", "offverb_V"],
            "offshld": ["offshld_ ", "offshld_S"],
            "forceuse": [
                "forceuse_ ",
                "forceuse_DO",
                "forceuse_DS",
                "forceuse_OR",
                "forceuse_OT",
                "forceuse_SF",
                "forceuse_SW",
            ],
            "haircolr": [
                "haircolr_BA",
                "haircolr_BK",
                "haircolr_BL",
                "haircolr_BR",
                "haircolr_DY",
                "haircolr_GY",
                "haircolr_RA",
                "haircolr_SN",
                "haircolr_SP",
                "haircolr_XX",
                "haircolr_ZZ",
            ],
            "eyecolor": [
                "eyecolor_BK",
                "eyecolor_BL",
                "eyecolor_BR",
                "eyecolor_DF",
                "eyecolor_GR",
                "eyecolor_GY",
                "eyecolor_HA",
                "eyecolor_XX",
                "eyecolor_Z",
            ],
            "build": ["build_H", "build_M", "build_T", "build_U", "build_Z"],
        }
        discrete_features = flatten_dict(disc_feature_groups)

        continuous_features = ["perstop", "ht_feet", "age", "ht_inch", "perobs", "weight"]

        sens_attr_spec: Union[str, LabelSpec]
        if self.split is SqfSplits.SEX:
            sens_attr_spec = "sex"
            s_prefix = ["sex"]
            class_label_spec = "weapon"
            class_label_prefix = ["weapon"]
        elif self.split is SqfSplits.RACE:
            sens_attr_spec = "race"
            s_prefix = ["race"]
            class_label_spec = "weapon"
            class_label_prefix = ["weapon"]
        elif self.split is SqfSplits.RACE_SEX:
            sens_attr_spec = simple_spec({"sex": ["sex"], "race": ["race"]})
            s_prefix = ["race", "sex"]
            class_label_spec = "weapon"
            class_label_prefix = ["weapon"]
        elif self.split is SqfSplits.CUSTOM:
            sens_attr_spec = ""
            s_prefix = []
            class_label_spec = ""
            class_label_prefix = []
        else:
            raise NotImplementedError
        super().__init__(
            name=f"SQF {self.split.value}",
            num_samples=12347,
            filename_or_path="sqf.csv",
            features=discrete_features + continuous_features,
            cont_features=continuous_features,
            s_prefix=s_prefix,
            sens_attr_spec=sens_attr_spec,
            class_label_prefix=class_label_prefix,
            class_label_spec=class_label_spec,
            discrete_only=self.discrete_only,
            discrete_feature_groups=disc_feature_groups,
        )



"""Paper: "Achieving Equalized Odds by Resampling Sensitive Attributes," Y. Romano, S. Bates, and E. J. Candès, 2020.

https://github.com/yromano/fair_dummies
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import List
from typing_extensions import Literal, TypedDict

from ranzen import implements

from ethicml.algorithms.inprocess.in_algorithm import HyperParamType
from ethicml.algorithms.inprocess.in_subprocess import InAlgorithmSubprocess

__all__ = ["FairDummies", "FairDummiesArgs"]


class FairDummiesArgs(TypedDict):
    """Args for the Fair Dummies implementation."""

    lr: float
    pretrain_pred_epochs: int
    pretrain_dis_epochs: int
    epochs: int
    loss_steps: int
    dis_steps: int
    batch_size: int
    model_type: Literal["deep_model", "linear_model"]
    lambda_vec: float
    second_moment_scaling: float


@dataclass
class FairDummies(InAlgorithmSubprocess):
    """FairDummie Method for enforcing Eq.Odds."""

    lr: float = 1e-3
    pretrain_pred_epochs: int = 2
    pretrain_dis_epochs: int = 2
    epochs: int = 50
    loss_steps: int = 1
    dis_steps: int = 1
    batch_size: int = 32
    model_type: str = "deep_model"
    lambda_vec: float = 0.9
    second_moment_scaling: float = 1e-5

    @implements(InAlgorithmSubprocess)
    def _get_path_to_script(self) -> List[str]:
        return ["-m", "ethicml.implementations.fair_dummies_romano"]

    @implements(InAlgorithmSubprocess)
    def _get_flags(self) -> FairDummiesArgs:
        model_type: Literal["deep_model", "linear_model"] = (
            "deep_model" if self.model_type.lower() == "deep_model" else "linear_model"
        )
        return {
            "lr": self.lr,
            "pretrain_pred_epochs": self.pretrain_pred_epochs,
            "pretrain_dis_epochs": self.pretrain_dis_epochs,
            "epochs": self.epochs,
            "loss_steps": self.loss_steps,
            "dis_steps": self.dis_steps,
            "batch_size": self.batch_size,
            "model_type": model_type,
            "lambda_vec": self.lambda_vec,
            "second_moment_scaling": self.second_moment_scaling,
        }

    @implements(InAlgorithmSubprocess)
    def get_hyperparameters(self) -> HyperParamType:
        return {
            "lr": self.lr,
            "pretrain_pred_epochs": self.pretrain_pred_epochs,
            "pretrain_dis_epochs": self.pretrain_dis_epochs,
            "epochs": self.epochs,
            "loss_steps": self.loss_steps,
            "dis_steps": self.dis_steps,
            "batch_size": self.batch_size,
            "model_type": self.model_type,
            "lambda_vec": self.lambda_vec,
            "second_moment_scaling": self.second_moment_scaling,
        }

    @implements(InAlgorithmSubprocess)
    def get_name(self) -> str:
        return f"Fair Dummies {self.model_type}"



"""Class to describe the 'LSAC Law School Admissions' dataset.

This dataset comes from Kusner et al Counterfactual Fairness.

Link to repo: https://github.com/mkusner/counterfactual-fairness/


@inproceedings{NIPS2017_a486cd07,
 author = {Kusner, Matt J and Loftus, Joshua and Russell, Chris and Silva, Ricardo},
 booktitle = {Advances in Neural Information Processing Systems},
 editor = {I. Guyon and U. V. Luxburg and S. Bengio and H. Wallach and R. Fergus and S. Vishwanathan and R. Garnett},
 pages = {},
 publisher = {Curran Associates, Inc.},
 title = {Counterfactual Fairness},
 url = {https://proceedings.neurips.cc/paper/2017/file/a486cd07e4ac3d270571622f4f316ec5-Paper.pdf},
 volume = {30},
 year = {2017}
}
"""
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Mapping, Type, Union

from ..dataset import LoadableDataset
from ..util import LabelGroup, flatten_dict, simple_spec

__all__ = ["Law", "LawSplits"]


class LawSplits(Enum):
    """Splits for the Law dataset."""

    SEX = "Sex"
    RACE = "Race"
    SEX_RACE = "Sex-Race"
    CUSTOM = "Custom"


@dataclass
class Law(LoadableDataset):
    """LSAC Law School dataset."""

    split: LawSplits = LawSplits.SEX

    Splits: ClassVar[Type[LawSplits]] = LawSplits
    """Shorthand for the Enum that defines the splits associated with this class."""

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "Sex": ["Sex_1", "Sex_2"],
            "Race": [
                "Race_Amerindian",
                "Race_Asian",
                "Race_Black",
                "Race_Hispanic",
                "Race_Mexican",
                "Race_Other",
                "Race_Puertorican",
                "Race_White",
            ],
            "PF": ["PF_0", "PF_1"],
        }
        discrete_features = flatten_dict(disc_feature_groups)

        continuous_features = ["LSAT", "UGPA", "ZFYA"]

        if self.split is LawSplits.SEX:
            sens_attr_spec: Union[str, Mapping[str, LabelGroup]] = "Sex_1"
            s_prefix = ["Sex", "Race"]
            class_label_spec = "PF_1"
            class_label_prefix = ["PF"]
        elif self.split is LawSplits.RACE:
            sens_attr_spec = "Race_White"
            s_prefix = ["Race", "Sex"]
            class_label_spec = "PF_1"
            class_label_prefix = ["PF"]
        elif self.split is LawSplits.SEX_RACE:
            sens_attr_spec = simple_spec({"Sex": ["Sex_1"], "Race": disc_feature_groups["Race"]})
            s_prefix = ["Race", "Sex"]
            class_label_spec = "PF_1"
            class_label_prefix = ["PF"]
        elif self.split is LawSplits.CUSTOM:
            sens_attr_spec = ""
            s_prefix = []
            class_label_spec = ""
            class_label_prefix = []
        else:
            raise NotImplementedError

        name = f"Law {self.split.value}"

        super().__init__(
            name=name,
            num_samples=21_791,
            features=discrete_features + continuous_features,
            cont_features=continuous_features,
            sens_attr_spec=sens_attr_spec,
            class_label_spec=class_label_spec,
            filename_or_path="law.csv.zip",
            s_prefix=s_prefix,
            class_label_prefix=class_label_prefix,
            discrete_only=self.discrete_only,
            discrete_feature_groups=disc_feature_groups,
        )



"""Abstract Base Class of all algorithms in the framework."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import asdict
from typing import ClassVar, TypeVar
from typing_extensions import final

from ethicml.algorithms.algorithm_base import Algorithm
from ethicml.utility import DataTuple, HyperParamType, Prediction, TestTuple

__all__ = [
    "InAlgorithm",
    "InAlgorithmDC",
    "InAlgorithmNoParams",
]

_I = TypeVar("_I", bound="InAlgorithm")


class InAlgorithm(Algorithm, ABC):
    """Abstract Base Class for algorithms that run in the middle of the pipeline."""

    is_fairness_algo: ClassVar[bool] = True  # should be overwritten by subclasses

    @abstractmethod
    def fit(self: _I, train: DataTuple, seed: int = 888) -> _I:
        """Fit Algorithm on the given data.

        :param train: Data tuple of the training data.
        :param seed: Random seed for model initialization.
        :returns: Self, but trained.
        """

    @abstractmethod
    def predict(self, test: TestTuple) -> Prediction:
        """Make predictions on the given data.

        :param test: Data to evaluate on.
        :returns: Predictions on the test data.
        """

    @abstractmethod
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        """Run Algorithm on the given data.

        :param train: Data tuple of the training data.
        :param test: Data to evaluate on.
        :param seed: Random seed for model initialization.
        :returns: Predictions on the test data.
        """

    @final
    def run_test(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        """Run with reduced training set so that it finishes quicker.

        :param train: Data tuple of the training data.
        :param test: Data to evaluate on.
        :param seed: Random seed for model initialization.
        :returns: Predictions on the test data.
        """
        train_testing = train.get_n_samples()
        return self.run(train_testing, test, seed)

    @abstractmethod
    def get_hyperparameters(self) -> HyperParamType:
        """Return list of hyperparameters."""


# ======== base classes with different default implementations of ``hyperparameters`` =========
class InAlgorithmNoParams(InAlgorithm, ABC):
    """Base class for algorithms without parameters."""

    @final
    def get_hyperparameters(self) -> HyperParamType:
        """Return list of hyperparameters."""
        return {}


class InAlgorithmDC(InAlgorithm, ABC):
    """Base class for algorithms that are dataclasses."""

    @final
    def get_hyperparameters(self) -> HyperParamType:
        """Return list of hyperparameters."""
        return asdict(self)



"""Module for kind of useful things that don't really belong anywhere else (just yet)."""

from .activation import *
from .data_helpers import *
from .data_structures import *
from .heaviside import *



"""Class to describe features of the German dataset."""
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type

from ..dataset import LoadableDataset
from ..util import flatten_dict

__all__ = ["German", "GermanSplits"]


class GermanSplits(Enum):
    """Splits for the German dataset."""

    SEX = "Sex"
    CUSTOM = "Custom"


@dataclass
class German(LoadableDataset):
    """German credit dataset."""

    split: GermanSplits = GermanSplits.SEX

    Splits: ClassVar[Type[GermanSplits]] = GermanSplits
    """Shorthand for the Enum that defines the splits associated with this class."""

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "sex": ["sex"],
            "age": ["age"],
            "credit-label": ["credit-label"],
            "status": ["status_A11", "status_A12", "status_A13", "status_A14"],
            "credit-history": [
                "credit-history_A30",
                "credit-history_A31",
                "credit-history_A32",
                "credit-history_A33",
                "credit-history_A34",
            ],
            "purpose": [
                "purpose_A40",
                "purpose_A41",
                "purpose_A410",
                "purpose_A42",
                "purpose_A43",
                "purpose_A44",
                "purpose_A45",
                "purpose_A46",
                "purpose_A48",
                "purpose_A49",
            ],
            "savings": [
                "savings_A61",
                "savings_A62",
                "savings_A63",
                "savings_A64",
                "savings_A65",
            ],
            "employment": [
                "employment_A71",
                "employment_A72",
                "employment_A73",
                "employment_A74",
                "employment_A75",
            ],
            "other-debtors": [
                "other-debtors_A101",
                "other-debtors_A102",
                "other-debtors_A103",
            ],
            "property": [
                "property_A121",
                "property_A122",
                "property_A123",
                "property_A124",
            ],
            "installment-plans": [
                "installment-plans_A141",
                "installment-plans_A142",
                "installment-plans_A143",
            ],
            "housing": [
                "housing_A151",
                "housing_A152",
                "housing_A153",
            ],
            "skill-level": [
                "skill-level_A171",
                "skill-level_A172",
                "skill-level_A173",
                "skill-level_A174",
            ],
            "telephone": [
                "telephone_A191",
                "telephone_A192",
            ],
            "foreign-worker": [
                "foreign-worker_A201",
                "foreign-worker_A202",
            ],
        }
        discrete_features = flatten_dict(disc_feature_groups)

        continuous_features = [
            "month",
            "credit-amount",
            "investment-as-income-percentage",
            "residence-since",
            "number-of-credits",
            "people-liable-for",
        ]

        if self.split is GermanSplits.SEX:
            sens_attr_spec = "sex"
            s_prefix = ["sex"]
            class_label_spec = "credit-label"
            class_label_prefix = ["credit-label"]
        elif self.split is GermanSplits.CUSTOM:
            sens_attr_spec = ""
            s_prefix = []
            class_label_spec = ""
            class_label_prefix = []
        else:
            raise NotImplementedError

        super().__init__(
            name=f"German {self.split.value}",
            num_samples=1_000,
            filename_or_path="german.csv",
            features=discrete_features + continuous_features,
            cont_features=continuous_features,
            s_prefix=s_prefix,
            sens_attr_spec=sens_attr_spec,
            class_label_prefix=class_label_prefix,
            class_label_spec=class_label_spec,
            discrete_only=self.discrete_only,
            discrete_feature_groups=disc_feature_groups,
        )



"""Class to describe features of the Heritage Health dataset."""
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type

from ..dataset import LoadableDataset
from ..util import flatten_dict

__all__ = ["Health", "HealthSplits"]


class HealthSplits(Enum):
    """Splits for the Health dataset."""

    SEX = "Sex"
    CUSTOM = "Custom"


@dataclass
class Health(LoadableDataset):
    """Heritage Health dataset."""

    split: HealthSplits = HealthSplits.SEX

    Splits: ClassVar[Type[HealthSplits]] = HealthSplits
    """Shorthand for the Enum that defines the splits associated with this class."""

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "age": [
                "age_05",
                "age_15",
                "age_25",
                "age_35",
                "age_45",
                "age_55",
                "age_65",
                "age_75",
                "age_85",
            ],
            "sex": [
                "sexMALE",
                "sexFEMALE",
            ],
            "labNull": ["labNull"],
            "drugNull": ["drugNull"],
            "ClaimsTruncated": ["ClaimsTruncated"],
            "Charlson>0": ["Charlson>0"],
        }
        discrete_features = flatten_dict(disc_feature_groups)

        continuous_features = [
            "no_Claims",
            "no_Providers",
            "no_Vendors",
            "no_PCPs",
            "no_PlaceSvcs",
            "no_Specialities",
            "no_PrimaryConditionGroups",
            "no_ProcedureGroups",
            "PayDelay_max",
            "PayDelay_min",
            "PayDelay_ave",
            "PayDelay_stdev",
            "LOS_max",
            "LOS_min",
            "LOS_ave",
            "LOS_stdev",
            "LOS_TOT_UNKNOWN",
            "LOS_TOT_SUPRESSED",
            "LOS_TOT_KNOWN",
            "dsfs_max",
            "dsfs_min",
            "dsfs_range",
            "dsfs_ave",
            "dsfs_stdev",
            "pcg1",
            "pcg2",
            "pcg3",
            "pcg4",
            "pcg5",
            "pcg6",
            "pcg7",
            "pcg8",
            "pcg9",
            "pcg10",
            "pcg11",
            "pcg12",
            "pcg13",
            "pcg14",
            "pcg15",
            "pcg16",
            "pcg17",
            "pcg18",
            "pcg19",
            "pcg20",
            "pcg21",
            "pcg22",
            "pcg23",
            "pcg24",
            "pcg25",
            "pcg26",
            "pcg27",
            "pcg28",
            "pcg29",
            "pcg30",
            "pcg31",
            "pcg32",
            "pcg33",
            "pcg34",
            "pcg35",
            "pcg36",
            "pcg37",
            "pcg38",
            "pcg39",
            "pcg40",
            "pcg41",
            "pcg42",
            "pcg43",
            "pcg44",
            "pcg45",
            "pcg46",
            "sp1",
            "sp2",
            "sp3",
            "sp4",
            "sp5",
            "sp6",
            "sp7",
            "sp8",
            "sp9",
            "sp10",
            "sp11",
            "sp12",
            "sp13",
            "pg1",
            "pg2",
            "pg3",
            "pg4",
            "pg5",
            "pg6",
            "pg7",
            "pg8",
            "pg9",
            "pg10",
            "pg11",
            "pg12",
            "pg13",
            "pg14",
            "pg15",
            "pg16",
            "pg17",
            "pg18",
            "ps1",
            "ps2",
            "ps3",
            "ps4",
            "ps5",
            "ps6",
            "ps7",
            "ps8",
            "ps9",
            "drugCount_max",
            "drugCount_min",
            "drugCount_ave",
            "drugcount_months",
            "labCount_max",
            "labCount_min",
            "labCount_ave",
            "labcount_months",
        ]

        if self.split is HealthSplits.SEX:
            sens_attr_spec = "sexMALE"
            s_prefix = ["sex"]
            class_label_spec = "Charlson>0"
            class_label_prefix = ["Charlson>0"]
        elif self.split is HealthSplits.CUSTOM:
            sens_attr_spec = ""
            s_prefix = []
            class_label_spec = ""
            class_label_prefix = []
        else:
            raise NotImplementedError
        super().__init__(
            name="Health",
            num_samples=171_067,
            filename_or_path="health.csv.zip",
            features=discrete_features + continuous_features,
            cont_features=continuous_features,
            s_prefix=s_prefix,
            sens_attr_spec=sens_attr_spec,
            class_label_prefix=class_label_prefix,
            class_label_spec=class_label_spec,
            discrete_only=self.discrete_only,
            discrete_feature_groups=disc_feature_groups,
        )



"""Classes related to running algorithms in subprocesses."""
import json
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, List, Mapping, TypeVar, Union
from typing_extensions import Literal, TypeAlias, TypedDict, final

from ethicml.algorithms.algorithm_base import SubprocessAlgorithmMixin
from ethicml.algorithms.inprocess.in_algorithm import InAlgorithm
from ethicml.utility.data_structures import DataTuple, Prediction, TestTuple

__all__ = ["InAlgoArgs", "InAlgorithmSubprocess"]


class InAlgoRunArgs(TypedDict):
    """Base arguments for the ``run`` function of subprocess in-process methods."""

    mode: Literal["run"]
    predictions: str  # path to where the predictions should be stored
    # paths to the files with the data
    train: str
    test: str
    seed: int


class InAlgoFitArgs(TypedDict):
    """Base arguments for the ``fit`` function of subprocess in-process methods."""

    mode: Literal["fit"]
    train: str
    model: str  # path to where the model weights are stored
    seed: int


class InAlgoPredArgs(TypedDict):
    """Base arguments for the ``predict`` function of subprocess in-process methods."""

    mode: Literal["predict"]
    predictions: str
    test: str
    model: str


InAlgoArgs: TypeAlias = Union[InAlgoFitArgs, InAlgoPredArgs, InAlgoRunArgs]


_IS = TypeVar("_IS", bound="InAlgorithmSubprocess")


@dataclass  # type: ignore[misc]  # mypy doesn't allow abstract dataclasses because mypy is stupid
class InAlgorithmSubprocess(SubprocessAlgorithmMixin, InAlgorithm):
    """In-Algorithm that uses a subprocess to run.

    :param dir: Directory to store the model.
    """

    dir: Path = Path(".")

    @property
    @final
    def model_path(self) -> Path:
        """Path to where the model with be stored."""
        return self.dir.resolve(strict=True) / f"model_{self.name}.joblib"

    @final
    def fit(self: _IS, train: DataTuple, seed: int = 888) -> _IS:
        """Fit Algorithm in a subprocess on the given data.

        :param train: Data tuple of the training data.
        :param seed: Random seed for model initialization.
        :returns: Self, but trained.
        """
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            train_path = tmp_path / "train.npz"
            train.to_npz(train_path)
            args: InAlgoFitArgs = {
                "mode": "fit",
                "train": str(train_path),
                "model": str(self.model_path),
                "seed": seed,
            }
            self.call_script(self._script_command(args))
            return self

    @final
    def predict(self, test: TestTuple) -> Prediction:
        """Make predictions in a subprocess on the given data.

        :param test: Data to evaluate on.
        :returns: Predictions on the test data.
        """
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            test_path = tmp_path / "test.npz"
            pred_path = tmp_path / "predictions.npz"
            test.to_npz(test_path)
            args: InAlgoPredArgs = {
                "mode": "predict",
                "test": str(test_path),
                "predictions": str(pred_path),
                "model": str(self.model_path),
            }
            self.call_script(self._script_command(args))
            return Prediction.from_npz(pred_path)

    @final
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        """Run Algorithm in a subprocess on the given data.

        :param train: Data tuple of the training data.
        :param test: Data to evaluate on.
        :param seed: Random seed for model initialization.
        :returns: Predictions on the test data.
        """
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            train_path = tmp_path / "train.npz"
            test_path = tmp_path / "test.npz"
            pred_path = tmp_path / "predictions.npz"
            train.to_npz(train_path)
            test.to_npz(test_path)
            args: InAlgoRunArgs = {
                "mode": "run",
                "train": str(train_path),
                "test": str(test_path),
                "predictions": str(pred_path),
                "seed": seed,
            }
            self.call_script(self._script_command(args))
            return Prediction.from_npz(pred_path)

    @final
    def _script_command(self, in_algo_args: InAlgoArgs) -> List[str]:
        """Return the command that will run the script.

        The flag interface consists of two strings, both JSON strings: the general in-algo flags
        and then the more specific flags for the algorithm.

        :param in_algo_args: Arguments for the script.
        :returns: List of strings that will be passed to ``subprocess.run``.
        """
        interface = [
            json.dumps(in_algo_args, separators=(',', ':')),
            json.dumps(self._get_flags(), separators=(',', ':')),
        ]
        return self._get_path_to_script() + interface

    @abstractmethod
    def _get_path_to_script(self) -> List[str]:
        """Return arguments that are passed to the python executable."""

    @abstractmethod
    def _get_flags(self) -> Mapping[str, Any]:
        """Return flags that are used to configure this algorithm."""



"""EthicML."""
from .algorithms.algorithm_base import *
from .algorithms.inprocess.adv_debiasing import *
from .algorithms.inprocess.agarwal_reductions import *
from .algorithms.inprocess.blind import *
from .algorithms.inprocess.fair_dummies import *
from .algorithms.inprocess.fairness_wo_demographics import *
from .algorithms.inprocess.hgr import *
from .algorithms.inprocess.in_algorithm import *
from .algorithms.inprocess.in_subprocess import *
from .algorithms.inprocess.installed_model import *
from .algorithms.inprocess.kamiran import *
from .algorithms.inprocess.kamishima import *
from .algorithms.inprocess.logistic_regression import *
from .algorithms.inprocess.majority import *
from .algorithms.inprocess.manual import *
from .algorithms.inprocess.mlp import *
from .algorithms.inprocess.oracle import *
from .algorithms.inprocess.svm import *
from .algorithms.inprocess.zafar import *
from .algorithms.postprocess.dp_flip import *
from .algorithms.postprocess.hardt import *
from .algorithms.postprocess.post_algorithm import *
from .algorithms.preprocess.beutel import *
from .algorithms.preprocess.calders import *
from .algorithms.preprocess.pre_algorithm import *
from .algorithms.preprocess.pre_subprocess import *
from .algorithms.preprocess.upsampler import *
from .algorithms.preprocess.vfae import *
from .algorithms.preprocess.zemel import *
from .common import *
from .data.dataset import *
from .data.load import *
from .data.lookup import *
from .data.tabular_data.acs import *
from .data.tabular_data.admissions import *
from .data.tabular_data.adult import *
from .data.tabular_data.compas import *
from .data.tabular_data.credit import *
from .data.tabular_data.crime import *
from .data.tabular_data.german import *
from .data.tabular_data.health import *
from .data.tabular_data.law import *
from .data.tabular_data.lipton import *
from .data.tabular_data.non_binary_toy import *
from .data.tabular_data.nursery import *
from .data.tabular_data.sqf import *
from .data.tabular_data.synthetic import *
from .data.tabular_data.toy import *
from .data.util import *
from .evaluators.cross_validator import *
from .evaluators.evaluate_models import *
from .evaluators.parallelism import *
from .metrics import *
from .preprocessing.adjust_labels import *
from .preprocessing.biased_split import *
from .preprocessing.domain_adaptation import *
from .preprocessing.feature_binning import *
from .preprocessing.scaling import *
from .preprocessing.splits import *
from .utility.activation import *
from .utility.data_helpers import *
from .utility.data_structures import *
from .utility.heaviside import *
from .visualisation.plot import *



"""Add column names to the CSV from Harvard dataverse 'UFRGS Entrance Exam and GPA Data'.

Persistent link: https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/O35FW8

Data Description: Entrance exam scores of students applying to a university in Brazil
(Federal University of Rio Grande do Sul), along with the students' GPAs during the first three
semesters at university. In this dataset, each row contains anonymized information about an
applicant's scores on nine exams taken as part of the application process to the university, as
well as their corresponding GPA during the first three semesters at university.
The dataset has 43,303 rows, each corresponding to one student.
The columns correspond to:
1) Gender. 0 denotes female and 1 denotes male.
2) Score on physics exam
3) Score on biology exam
4) Score on history exam
5) Score on second language exam
6) Score on geography exam
7) Score on literature exam
8) Score on Portuguese essay exam
9) Score on math exam
10) Score on chemistry exam
11) Mean GPA during first three semesters at university, on a 4.0 scale.

We replace the mean GPA with a binary label Y representing whether the student’s GPA was above 3.0


```bibtex
@data{DVN/O35FW8_2019,
  author    = {Castro~da~Silva, Bruno},
  publisher = {Harvard Dataverse},
  title     = {{UFRGS Entrance Exam and GPA Data}},
  UNF       = {UNF:6:MQqEQGXiIfQTbS7q9QJ5uw==},
  year      = {2019},
  version   = {V1},
  doi       = {10.7910/DVN/O35FW8},
  url       = {https://doi.org/10.7910/DVN/O35FW8}
}
```
"""


import pandas as pd


def run_generate_admissions() -> None:
    """Add a the header solumns, binarise label and shuffle."""
    data = pd.read_csv("raw/ufrgs_entrance_exam_and_gpa.csv.zip", header=None)

    # Give data column names
    columns = [
        "gender",
        "physics",
        "biology",
        "history",
        "language",
        "geography",
        "literature",
        "essay",
        "math",
        "chemistry",
        "gpa",
    ]
    data.columns = pd.Index(columns)

    # Binarize the GPA column
    data["gpa"] = data["gpa"] >= 3.0
    data["gpa"] = data["gpa"].astype("int")

    # Shuffle the data
    data = data.sample(frac=1.0, random_state=888).reset_index(drop=True)

    # Save the CSV
    compression_opts = dict(method='zip', archive_name='admissions.csv')
    data.to_csv("./admissions.csv.zip", index=False, compression=compression_opts)


if __name__ == '__main__':
    run_generate_admissions()



"""DRO Loss."""
from typing import Optional, Type

from ranzen import implements
from torch import Tensor, nn
from torch.nn.modules.loss import NLLLoss, _Loss

__all__ = ["DROLoss"]


class DROLoss(nn.Module):
    """Fairness Without Demographics Loss."""

    def __init__(self, loss_module: Optional[Type[_Loss]] = None, eta: float = 0.5):
        """Set up the loss, set which loss you want to optimize and the eta to offset by."""
        super().__init__()
        if loss_module is None:
            loss_module = NLLLoss
        self.loss = loss_module(reduction="none")
        self.eta = eta

    @implements(nn.Module)
    def forward(self, pred: Tensor, target: Tensor) -> Tensor:  # pylint: disable=arguments-differ
        return (self.loss(pred, target=target) - self.eta).relu().pow(2).mean()



"""Zemel's Learned Fair Representations."""
from dataclasses import dataclass
from typing import List, Optional
from typing_extensions import TypedDict

from ranzen import implements

from .pre_subprocess import PreAlgorithmSubprocess

__all__ = ["Zemel"]


class ZemelArgs(TypedDict):
    """Arguments for the Zemel algorithm."""

    clusters: int
    Ax: float
    Ay: float
    Az: float
    max_iter: int
    maxfun: int
    epsilon: float
    threshold: float


@dataclass
class Zemel(PreAlgorithmSubprocess):
    """AIF360 implementation of Zemel's LFR."""

    threshold: float = 0.5
    clusters: int = 2
    Ax: float = 0.01
    Ay: float = 0.1
    Az: float = 0.5
    max_iter: int = 5_000
    maxfun: int = 5_000
    epsilon: float = 1e-5

    def __post_init__(self) -> None:
        self._in_size: Optional[int] = None  # the super class will set this for us

    @implements(PreAlgorithmSubprocess)
    def _get_flags(self) -> ZemelArgs:
        return {
            "clusters": self.clusters,
            "Ax": self.Ax,
            "Ay": self.Ay,
            "Az": self.Az,
            "max_iter": self.max_iter,
            "maxfun": self.maxfun,
            "epsilon": self.epsilon,
            "threshold": self.threshold,
        }

    @implements(PreAlgorithmSubprocess)
    def get_name(self) -> str:
        return "Zemel"

    @implements(PreAlgorithmSubprocess)
    def get_out_size(self) -> int:
        assert self._in_size is not None
        return self._in_size

    @implements(PreAlgorithmSubprocess)
    def _get_path_to_script(self) -> List[str]:
        return ["-m", "ethicml.implementations.zemel"]



"""Abstract Base Class of all post-processing algorithms in the framework."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeVar

from ethicml.utility import DataTuple, Prediction, TestTuple

from ..algorithm_base import Algorithm

__all__ = ["PostAlgorithm"]

_PA = TypeVar("_PA", bound="PostAlgorithm")


class PostAlgorithm(Algorithm, ABC):
    """Abstract Base Class for all algorithms that do post-processing."""

    @abstractmethod
    def fit(self: _PA, train_predictions: Prediction, train: DataTuple) -> _PA:
        """Run Algorithm on the given data.

        :param train_predictions: Predictions on a training set.
        :param train: The training set with the correct labels and sensitive attributes.
        :returns: Self.
        """

    @abstractmethod
    def predict(self, test_predictions: Prediction, test: TestTuple, seed: int = 888) -> Prediction:
        """Run Algorithm on the given data.

        :param test_predictions: Predictions on a test set.
        :param test: The test set with the sensitive attributes.
        :returns: Post-processed predictions on the test set.
        """

    @abstractmethod
    def run(
        self,
        train_predictions: Prediction,
        train: DataTuple,
        test_predictions: Prediction,
        test: TestTuple,
        seed: int = 888,
    ) -> Prediction:
        """Make predictions fair.

        :param train_predictions: Predictions on a training set.
        :param train: The training set with the correct labels and sensitive attributes.
        :param test_predictions: Predictions on the test set.
        :param test: The test set with the sensitive attributes.
        :returns: Post-processed predictions on the test set.
        """



"""Transparently show how the UCI Nursey dataset was modified from the raw download."""
# The Heritage Health dataset. It needs some (mild) preprocessing before we can plug and play.

import pandas as pd

if __name__ == "__main__":

    df = pd.read_csv(
        "raw/nursery_raw.csv",
        names=[
            "parents",
            "has_nurs",
            "form",
            "children",
            "housing",
            "finance",
            "social",
            "health",
            "class",
        ],
    )

    # Drop rows
    # indexNames = df[df['class'] == 'recommend'].index
    # df.drop(indexNames, inplace=True)

    # Add binary class column
    features = ["children"]
    for column in features:
        df[column] = df[column].astype("category").cat.codes

    features1 = ['form', 'health', "finance", "class", "parents", "has_nurs", "housing", "social"]
    x1 = df.drop(features1, axis=1)

    x2 = df.drop(features, axis=1)
    x2 = pd.get_dummies(x2)

    df = pd.concat([x1, x2], axis=1)

    # Shuffle the data
    df = df.sample(frac=1.0, random_state=888).reset_index(drop=True)

    # Save the CSV
    compression_opts = dict(method='zip', archive_name='nursery.csv')
    df.to_csv("./nursery.csv.zip", index=False, compression=compression_opts)



"""Helper functions for working with DataFrames (and Series)."""
from typing import Optional, Union, overload

import pandas as pd

__all__ = ["undo_one_hot", "shuffle_df"]


def shuffle_df(df: pd.DataFrame, random_state: int) -> pd.DataFrame:
    """Shuffle a given dataframe.

    :param df:
    :param random_state:
    """
    return df.sample(frac=1.0, random_state=random_state).reset_index(drop=True)  # type: ignore[return-value]


@overload
def undo_one_hot(df: pd.DataFrame, new_column_name: None = None) -> pd.Series:
    ...


@overload
def undo_one_hot(df: pd.DataFrame, new_column_name: str) -> pd.DataFrame:
    ...


def undo_one_hot(
    df: pd.DataFrame, new_column_name: Optional[str] = None
) -> Union[pd.Series, pd.DataFrame]:
    """Undo one-hot encoding.

    :param df:
    :param new_column_name:  (Default: None)
    """
    # we have to overwrite the column names because `idxmax` uses the column names
    df.columns = pd.Index(range(df.shape[1]))
    result = df.idxmax(axis="columns")
    if new_column_name is not None:
        return result.to_frame(name=new_column_name)
    else:
        return result



"""Wrapper around Sci-Kit Learn Logistic Regression."""
from dataclasses import dataclass, field
from typing import ClassVar

import numpy as np
import pandas as pd
from ranzen import implements
from sklearn.linear_model import LogisticRegression, LogisticRegressionCV
from sklearn.model_selection import KFold

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmDC
from ethicml.utility import DataTuple, Prediction, SoftPrediction, TestTuple

__all__ = ["LR", "LRCV"]


@dataclass
class LR(InAlgorithmDC):
    """Logistic regression with hard predictions.

    This is a wrapper around Sci-Kit Learn's LogisticRegression.
    See `the documentation <https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html>`_
    for details.

    :param C: The regularization parameter.
    """

    is_fairness_algo: ClassVar[bool] = False
    C: float = field(default_factory=lambda: LogisticRegression().C)

    @implements(InAlgorithmDC)
    def get_name(self) -> str:
        return f"Logistic Regression (C={self.C})"

    @implements(InAlgorithmDC)
    def fit(self, train: DataTuple, seed: int = 888) -> InAlgorithmDC:
        random_state = np.random.RandomState(seed=seed)
        self.clf = LogisticRegression(
            solver="liblinear", random_state=random_state, C=self.C, multi_class="auto"
        )
        self.clf.fit(train.x, train.y.to_numpy().ravel())
        return self

    @implements(InAlgorithmDC)
    def predict(self, test: TestTuple) -> Prediction:
        return Prediction(hard=pd.Series(self.clf.predict(test.x)))

    @implements(InAlgorithmDC)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        random_state = np.random.RandomState(seed=seed)
        clf = LogisticRegression(
            solver="liblinear", random_state=random_state, C=self.C, multi_class="auto"
        )
        clf.fit(train.x, train.y.to_numpy().ravel())
        return SoftPrediction(soft=clf.predict_proba(test.x), info=self.get_hyperparameters())


@dataclass
class LRCV(InAlgorithmDC):
    """Kind of a cheap hack for now, but gives a proper cross-valudeted LR.

    :param n_splits: The number of splits for the cross-validation.
    """

    is_fairness_algo: ClassVar[bool] = False
    n_splits: int = 3

    @implements(InAlgorithmDC)
    def get_name(self) -> str:
        return "LRCV"

    @implements(InAlgorithmDC)
    def fit(self, train: DataTuple, seed: int = 888) -> InAlgorithmDC:
        random_state = np.random.RandomState(seed=seed)
        folder = KFold(n_splits=self.n_splits, shuffle=True, random_state=random_state)
        self.clf = LogisticRegressionCV(
            cv=folder, n_jobs=-1, random_state=random_state, solver="liblinear", multi_class="auto"
        )
        self.clf.fit(train.x, train.y.to_numpy().ravel())
        return self

    @implements(InAlgorithmDC)
    def predict(self, test: TestTuple) -> Prediction:
        params = self.get_hyperparameters()
        params["C"] = self.clf.C_[0]
        return SoftPrediction(soft=self.clf.predict_proba(test.x), info=params)

    @implements(InAlgorithmDC)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        random_state = np.random.RandomState(seed=seed)
        folder = KFold(n_splits=self.n_splits, shuffle=True, random_state=random_state)
        clf = LogisticRegressionCV(
            cv=folder, n_jobs=-1, random_state=random_state, solver="liblinear", multi_class="auto"
        )
        clf.fit(train.x, train.y.to_numpy().ravel())
        params = self.get_hyperparameters()
        params["C"] = clf.C_[0]
        return SoftPrediction(soft=clf.predict_proba(test.x), info=params)
